#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""FCMA preprocessing."""

# Authors: Yida Wang
# (Intel Labs), 2017

import math
import time
import numpy as np
import logging
from scipy.stats.mstats import zscore
from mpi4py import MPI
from enum import Enum

from ..image import mask_images, multimask_images


logger = logging.getLogger(__name__)

__all__ = [
    "generate_epochs_info",
    "prepare_fcma_data",
    "prepare_mvpa_data",
    "prepare_searchlight_mvpa_data",
    "RandomType",
]


def _separate_epochs(activity_data, epoch_list):
    """ create data epoch by epoch

    Separate data into epochs of interest specified in epoch_list
    and z-score them for computing correlation

    Parameters
    ----------
    activity_data: list of 2D array in shape [nVoxels, nTRs]
        the masked activity data organized in voxel*TR formats of all subjects
    epoch_list: list of 3D array in shape [condition, nEpochs, nTRs]
        specification of epochs and conditions
        assuming all subjects have the same number of epochs
        len(epoch_list) equals the number of subjects

    Returns
    -------
    raw_data: list of 2D array in shape [epoch length, nVoxels]
        the data organized in epochs
        and z-scored in preparation of correlation computation
        len(raw_data) equals the number of epochs
    labels: list of 1D array
        the condition labels of the epochs
        len(labels) labels equals the number of epochs
    """
    time1 = time.time()
    raw_data = []
    labels = []
    for sid in range(len(epoch_list)):
        epoch = epoch_list[sid]
        for cond in range(epoch.shape[0]):
            sub_epoch = epoch[cond, :, :]
            for eid in range(epoch.shape[1]):
                r = np.sum(sub_epoch[eid, :])
                if r > 0:   # there is an epoch in this condition
                    # mat is row-major
                    # regardless of the order of acitvity_data[sid]
                    mat = activity_data[sid][:, sub_epoch[eid, :] == 1]
                    mat = np.ascontiguousarray(mat.T)
                    mat = zscore(mat, axis=0, ddof=0)
                    # if zscore fails (standard deviation is zero),
                    # set all values to be zero
                    mat = np.nan_to_num(mat)
                    mat = mat / math.sqrt(r)
                    raw_data.append(mat)
                    labels.append(cond)
    time2 = time.time()
    logger.debug(
        'epoch separation done, takes %.2f s' %
        (time2 - time1)
    )
    return raw_data, labels


def _randomize_single_subject(data, seed=None):
    """Randomly permute the voxels of the subject.

     The subject is organized as Voxel x TR,
     this method shuffles the voxel dimension in place.

    Parameters
    ----------
    data: 2D array in shape [nVoxels, nTRs]
        Activity image data to be shuffled.
    seed: Optional[int]
        Seed for random state used implicitly for shuffling.

    Returns
    -------
    None.
    """
    if seed is not None:
        np.random.seed(seed)
    np.random.shuffle(data)


def _randomize_subject_list(data_list, random):
    """Randomly permute the voxels of a subject list.

     The method shuffles the subject one by one in place according to
     the random type. If RandomType.NORANDOM, return the original list.

    Parameters
    ----------
    data_list: list of 2D array in shape [nVxels, nTRs]
        Activity image data list to be shuffled.
    random: RandomType
        Randomization type.

    Returns
    -------
    None.
    """
    if random == RandomType.REPRODUCIBLE:
        for i in range(len(data_list)):
            _randomize_single_subject(data_list[i], seed=i)
    elif random == RandomType.UNREPRODUCIBLE:
        for data in data_list:
            _randomize_single_subject(data)


class RandomType(Enum):
    """Define the random types as enumeration

    NORANDOM means do not randomize the image data;
    REPRODUCIBLE means randomize the image data with a fixed seed so that the
    permutation holds between different runs;
    UNREPRODUCIBLE means truly randomize the image data which returns
    different results in different runs.
    """
    NORANDOM = 0
    REPRODUCIBLE = 1
    UNREPRODUCIBLE = 2


def prepare_fcma_data(images, conditions, mask1, mask2=None,
                      random=RandomType.NORANDOM, comm=MPI.COMM_WORLD):
    """Prepare data for correlation-based computation and analysis.

    Generate epochs of interests, then broadcast to all workers.

    Parameters
    ----------
    images: Iterable[SpatialImage]
        Data.
    conditions: List[UniqueLabelConditionSpec]
        Condition specification.
    mask1: np.ndarray
        Mask to apply to each image.
    mask2: Optional[np.ndarray]
        Mask to apply to each image.
        If it is not specified, the method will assign None to the returning
        variable raw_data2 and the self-correlation on raw_data1 will be
        computed
    random: Optional[RandomType]
        Randomize the image data within subject or not.
    comm: MPI.Comm
        MPI communicator to use for MPI operations.

    Returns
    -------
    raw_data1: list of 2D array in shape [epoch length, nVoxels]
        the data organized in epochs, specified by the first mask.
        len(raw_data) equals the number of epochs
    raw_data2: Optional, list of 2D array in shape [epoch length, nVoxels]
        the data organized in epochs, specified by the second mask if any.
        len(raw_data2) equals the number of epochs
    labels: list of 1D array
        the condition labels of the epochs
        len(labels) labels equals the number of epochs
    """
    rank = comm.Get_rank()
    labels = []
    raw_data1 = []
    raw_data2 = []
    if rank == 0:
        logger.info('start to apply masks and separate epochs')
        if mask2 is not None:
            masks = (mask1, mask2)
            activity_data1, activity_data2 = zip(*multimask_images(images,
                                                                   masks,
                                                                   np.float32))
            _randomize_subject_list(activity_data2, random)
            raw_data2, _ = _separate_epochs(activity_data2, conditions)
        else:
            activity_data1 = list(mask_images(images, mask1, np.float32))
        _randomize_subject_list(activity_data1, random)
        raw_data1, labels = _separate_epochs(activity_data1, conditions)
        time1 = time.time()
    raw_data_length = len(raw_data1)
    raw_data_length = comm.bcast(raw_data_length)
    # broadcast the data subject by subject to prevent size overflow
    for i in range(raw_data_length):
        if rank != 0:
            raw_data1.append(None)
            if mask2 is not None:
                raw_data2.append(None)
        raw_data1[i] = comm.bcast(raw_data1[i], root=0)
        if mask2 is not None:
            raw_data2[i] = comm.bcast(raw_data2[i], root=0)

    if comm.Get_size() > 1:
        labels = comm.bcast(labels, root=0)
        if rank == 0:
            time2 = time.time()
            logger.info(
                'data broadcasting done, takes %.2f s' %
                (time2 - time1)
            )
    if mask2 is None:
        raw_data2 = None
    return raw_data1, raw_data2, labels


def generate_epochs_info(epoch_list):
    """ use epoch_list to generate epoch_info defined below

    Parameters
    ----------
    epoch_list: list of 3D (binary) array in shape [condition, nEpochs, nTRs]
        Contains specification of epochs and conditions, assuming
        1. all subjects have the same number of epochs;
        2. len(epoch_list) equals the number of subjects;
        3. an epoch is always a continuous time course.

    Returns
    -------
    epoch_info: list of tuple (label, sid, start, end).
        label is the condition labels of the epochs;
        sid is the subject id, corresponding to the index of raw_data;
        start is the start TR of an epoch (inclusive);
        end is the end TR of an epoch(exclusive).
        Assuming len(labels) labels equals the number of epochs and
        the epochs of the same sid are adjacent in epoch_info
    """
    time1 = time.time()
    epoch_info = []
    for sid, epoch in enumerate(epoch_list):
        for cond in range(epoch.shape[0]):
            sub_epoch = epoch[cond, :, :]
            for eid in range(epoch.shape[1]):
                r = np.sum(sub_epoch[eid, :])
                if r > 0:   # there is an epoch in this condition
                    start = np.nonzero(sub_epoch[eid, :])[0][0]
                    epoch_info.append((cond, sid, start, start + r))
    time2 = time.time()
    logger.debug(
        'epoch separation done, takes %.2f s' %
        (time2 - time1)
    )
    return epoch_info


def prepare_mvpa_data(images, conditions, mask):
    """Prepare data for activity-based model training and prediction.

    Average the activity within epochs and z-scoring within subject.

    Parameters
    ----------
    images: Iterable[SpatialImage]
        Data.
    conditions: List[UniqueLabelConditionSpec]
        Condition specification.
    mask: np.ndarray
        Mask to apply to each image.

    Returns
    -------
    processed_data: 2D array in shape [num_voxels, num_epochs]
        averaged epoch by epoch processed data
    labels: 1D array
        contains labels of the data
    """
    activity_data = list(mask_images(images, mask, np.float32))
    epoch_info = generate_epochs_info(conditions)
    num_epochs = len(epoch_info)
    (d1, _) = activity_data[0].shape
    processed_data = np.empty([d1, num_epochs])
    labels = np.empty(num_epochs)
    subject_count = [0]  # counting the epochs per subject for z-scoring
    cur_sid = -1
    # averaging
    for idx, epoch in enumerate(epoch_info):
        labels[idx] = epoch[0]
        if cur_sid != epoch[1]:
            subject_count.append(0)
            cur_sid = epoch[1]
        subject_count[-1] += 1
        processed_data[:, idx] = \
            np.mean(activity_data[cur_sid][:, epoch[2]:epoch[3]],
                    axis=1)
    # z-scoring
    cur_epoch = 0
    for i in subject_count:
        if i > 1:
            processed_data[:, cur_epoch:cur_epoch + i] = \
                zscore(processed_data[:, cur_epoch:cur_epoch + i],
                       axis=1, ddof=0)
        cur_epoch += i
    # if zscore fails (standard deviation is zero),
    # set all values to be zero
    processed_data = np.nan_to_num(processed_data)

    return processed_data, labels


def prepare_searchlight_mvpa_data(images, conditions, data_type=np.float32,
                                  random=RandomType.NORANDOM):
    """ obtain the data for activity-based voxel selection using Searchlight

    Average the activity within epochs and z-scoring within subject,
    while maintaining the 3D brain structure. In order to save memory,
    the data is processed subject by subject instead of reading all in before
    processing. Assuming all subjects live in the identical cube.

    Parameters
    ----------
    images: Iterable[SpatialImage]
        Data.
    conditions: List[UniqueLabelConditionSpec]
        Condition specification.
    data_type
        Type to cast image to.
    random: Optional[RandomType]
        Randomize the image data within subject or not.

    Returns
    -------
    processed_data: 4D array in shape [brain 3D + epoch]
        averaged epoch by epoch processed data

    labels: 1D array
        contains labels of the data
    """
    time1 = time.time()
    epoch_info = generate_epochs_info(conditions)
    num_epochs = len(epoch_info)
    processed_data = None
    logger.info(
        'there are %d subjects, and in total %d epochs' %
        (len(conditions), num_epochs)
    )
    labels = np.empty(num_epochs)
    # assign labels
    for idx, epoch in enumerate(epoch_info):
        labels[idx] = epoch[0]
    # counting the epochs per subject for z-scoring
    subject_count = np.zeros(len(conditions), dtype=np.int32)

    logger.info('start to apply masks and separate epochs')
    for sid, f in enumerate(images):
        data = f.get_fdata().astype(data_type)
        [d1, d2, d3, d4] = data.shape
        if random == RandomType.REPRODUCIBLE:
            data = data.reshape((d1 * d2 * d3, d4))
            _randomize_single_subject(data, seed=sid)
            data = data.reshape((d1, d2, d3, d4))
        elif random == RandomType.UNREPRODUCIBLE:
            data = data.reshape((d1 * d2 * d3, d4))
            _randomize_single_subject(data)
            data = data.reshape((d1, d2, d3, d4))
        if processed_data is None:
            processed_data = np.empty([d1, d2, d3, num_epochs],
                                      dtype=data_type)
        # averaging
        for idx, epoch in enumerate(epoch_info):
            if sid == epoch[1]:
                subject_count[sid] += 1
                processed_data[:, :, :, idx] = \
                    np.mean(data[:, :, :, epoch[2]:epoch[3]], axis=3)

        logger.debug(
            'file %s is loaded and processed, with data shape %s',
            f.get_filename(), data.shape
        )
    # z-scoring
    cur_epoch = 0
    for i in subject_count:
        if i > 1:
            processed_data[:, :, :, cur_epoch:cur_epoch + i] = \
                zscore(processed_data[:, :, :, cur_epoch:cur_epoch + i],
                       axis=3, ddof=0)
        cur_epoch += i
    # if zscore fails (standard deviation is zero),
    # set all values to be zero
    processed_data = np.nan_to_num(processed_data)
    time2 = time.time()
    logger.info(
        'data processed for activity-based voxel selection, takes %.2f s' %
        (time2 - time1)
    )

    return processed_data, labels



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from brainiak.fcma.voxelselector import VoxelSelector
from brainiak.fcma.preprocessing import prepare_fcma_data
from brainiak import io
from sklearn import svm
import sys
from mpi4py import MPI
import logging
import numpy as np
import nibabel as nib

format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
# if want to output log to a file instead of outputting log to the console,
# replace "stream=sys.stdout" with "filename='fcma.log'"
logging.basicConfig(level=logging.INFO, format=format, stream=sys.stdout)
logger = logging.getLogger(__name__)

"""
example running command in run_voxel_selection.sh
"""
if __name__ == '__main__':
    if MPI.COMM_WORLD.Get_rank()==0:
        logger.info(
            'programming starts in %d process(es)' %
            MPI.COMM_WORLD.Get_size()
        )
    if len(sys.argv) != 7:
        logger.error('the number of input argument is not correct')
        sys.exit(1)
    data_dir = sys.argv[1]
    suffix = sys.argv[2]
    mask_file = sys.argv[3]
    epoch_file = sys.argv[4]
    images = io.load_images_from_dir(data_dir, suffix=suffix)
    mask = io.load_boolean_mask(mask_file)
    conditions = io.load_labels(epoch_file)
    raw_data, _, labels = prepare_fcma_data(images, conditions, mask)

    # setting the random argument produces random voxel selection results
    # for non-parametric statistical analysis.
    # There are three random options:
    # RandomType.NORANDOM is the default
    # RandomType.REPRODUCIBLE permutes the voxels in the same way every run
    # RandomType.UNREPRODUCIBLE permutes the voxels differently across runs
    # example:
    # from brainiak.fcma.preprocessing import RandomType
    # raw_data, _, labels = prepare_fcma_data(images, conditions, mask,
    #                                         random=RandomType.REPRODUCIBLE)

    # if providing two masks, just append the second mask as the last input argument
    # and specify raw_data2
    # example:
    # images = io.load_images_from_dir(data_dir, extension)
    # mask2 = io.load_boolean_mask('face_scene/mask.nii.gz')
    # raw_data, raw_data2, labels = prepare_fcma_data(images, conditions, mask,
    #                                                 mask2)

    epochs_per_subj = int(sys.argv[5])
    num_subjs = int(sys.argv[6])
    # the following line is an example to leaving a subject out
    #vs = VoxelSelector(labels[0:204], epochs_per_subj, num_subjs-1, raw_data[0:204])
    # if using all subjects
    vs = VoxelSelector(labels, epochs_per_subj, num_subjs, raw_data)
    # if providing two masks, just append raw_data2 as the last input argument
    #vs = VoxelSelector(labels, epochs_per_subj, num_subjs, raw_data, raw_data2=raw_data2)
    # for cross validation, use SVM with precomputed kernel
    clf = svm.SVC(kernel='precomputed', shrinking=False, C=10, gamma='auto')
    results = vs.run(clf)
    # this output is just for result checking
    if MPI.COMM_WORLD.Get_rank()==0:
        logger.info(
            'correlation-based voxel selection is done'
        )
        #print(results[0:100])
        mask_img = nib.load(mask_file)
        mask = mask_img.get_fdata().astype(bool)
        score_volume = np.zeros(mask.shape, dtype=np.float32)
        score = np.zeros(len(results), dtype=np.float32)
        seq_volume = np.zeros(mask.shape, dtype=int)
        seq = np.zeros(len(results), dtype=int)
        with open('result_list.txt', 'w') as fp:
            for idx, tuple in enumerate(results):
                fp.write(str(tuple[0]) + ' ' + str(tuple[1]) + '\n')
                score[tuple[0]] = tuple[1]
                seq[tuple[0]] = idx
        score_volume[mask] = score
        seq_volume[mask] = seq
        io.save_as_nifti_file(score_volume, mask_img.affine,
                              'result_score.nii.gz')
        io.save_as_nifti_file(seq_volume, mask_img.affine,
                                   'result_seq.nii.gz')



#  Copyright 2017 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Generic image functionality."""

__all__ = [
    "ConditionSpec",
    "MaskedMultiSubjectData",
    "mask_image",
    "mask_images",
    "multimask_images",
    "SingleConditionSpec",
]

import itertools

from typing import Optional, Iterable, Sequence, Type, TypeVar

import numpy as np

from nibabel.spatialimages import SpatialImage


T = TypeVar("T", bound="MaskedMultiSubjectData")


class MaskedMultiSubjectData(np.ndarray):
    """Array with shape n_TRs, n_voxels, n_subjects."""
    @classmethod
    def from_masked_images(cls: Type[T], masked_images: Iterable[np.ndarray],
                           n_subjects: int) -> T:
        """Create a new instance of MaskedMultiSubjecData from masked images.

        Parameters
        ----------
        masked_images : iterator
            Images from multiple subjects to stack along 3rd dimension
        n_subjects : int
            Number of subjects; must match the number of images

        Returns
        -------
        T
            A new instance of MaskedMultiSubjectData

        Raises
        ------
        ValueError
            Images have different shapes.

            The number of images differs from n_subjects.
        """
        images_iterator = iter(masked_images)
        first_image = next(images_iterator)
        first_image_shape = first_image.T.shape
        result = np.empty((first_image_shape[0], first_image_shape[1],
                           n_subjects))
        for n_images, image in enumerate(itertools.chain([first_image],
                                                         images_iterator)):
            image = image.T
            if image.shape != first_image_shape:
                raise ValueError("Image {} has different shape from first "
                                 "image: {} != {}".format(n_images,
                                                          image.shape,
                                                          first_image_shape))
            result[:, :, n_images] = image
        n_images += 1
        if n_images != n_subjects:
            raise ValueError("n_subjects != number of images: {} != {}"
                             .format(n_subjects, n_images))
        return result.view(cls)


class ConditionSpec(np.ndarray):
    """One-hot representation of conditions across epochs and TRs.

    The shape is (n_conditions, n_epochs, n_trs).
    """


class SingleConditionSpec(ConditionSpec):
    """ConditionSpec with a single condition applicable to an epoch."""

    def extract_labels(self) -> np.ndarray:
        """Extract condition labels.

        Returns
        -------
        np.ndarray
            The condition label of each epoch.
        """
        condition_idxs, epoch_idxs, _ = np.where(self)
        _, unique_epoch_idxs = np.unique(epoch_idxs, return_index=True)
        return condition_idxs[unique_epoch_idxs]


def mask_image(
        image: SpatialImage,
        mask: np.ndarray,
        data_type: Optional[type] = None
        ) -> np.ndarray:
    """Mask image after optionally casting its type.

    Parameters
    ----------
    image
        Image to mask. Can include time as the last dimension.
    mask
        Mask to apply. Must have the same shape as the image data.
    data_type
        Type to cast image to.

    Returns
    -------
    np.ndarray
        Masked image.

    Raises
    ------
    ValueError
        Image data and masks have different shapes.
    """
    image_data = image.get_fdata()
    if image_data.shape[:3] != mask.shape:
        raise ValueError("Image data and mask have different shapes.")
    if data_type is not None:
        cast_data: np.ndarray = image_data.astype(data_type)
    else:
        cast_data = image_data
    return cast_data[mask]


def multimask_images(
        images: Iterable[SpatialImage],
        masks: Sequence[np.ndarray],
        image_type: Optional[type] = None
        ) -> Iterable[Sequence[np.ndarray]]:
    """Mask images with multiple masks.

    Parameters
    ----------
    images:
        Images to mask.
    masks:
        Masks to apply.
    image_type:
        Type to cast images to.

    Yields
    ------
    Sequence[np.ndarray]
        For each mask, a masked image.
    """
    for image in images:
        yield [mask_image(image, mask, image_type) for mask in masks]


def mask_images(images: Iterable[SpatialImage], mask: np.ndarray,
                image_type: Optional[type] = None) -> Iterable[np.ndarray]:
    """Mask images.

    Parameters
    ----------
    images:
        Images to mask.
    mask:
        Mask to apply.
    image_type:
        Type to cast images to.

    Yields
    ------
    np.ndarray
        Masked image.
    """
    for imgs in multimask_images(images, (mask,), image_type):  # type: ignore
        yield imgs[0]  # type: ignore



from brainiak.matnormal.mnrsa import MNRSA
from brainiak.utils.utils import cov2corr
from brainiak.matnormal.covs import CovIdentity, CovDiagonal
from scipy.stats import norm
from brainiak.matnormal.utils import rmn
import numpy as np


def gen_U_nips2016_example():

    n_C = 16
    U = np.zeros([n_C, n_C])
    U = np.eye(n_C) * 0.6
    U[8:12, 8:12] = 0.8
    for cond in range(8, 12):
        U[cond, cond] = 1

    return U


def gen_brsa_data_matnorm_model(U, n_T, n_V, space_cov, time_cov, n_nureg):

    n_C = U.shape[0]
    beta = rmn(U, space_cov)
    X = rmn(np.eye(n_T), np.eye(n_C))
    beta_0 = rmn(np.eye(n_nureg), space_cov)
    X_0 = rmn(np.eye(n_T), np.eye(n_nureg))
    Y_hat = X.dot(beta) + X_0.dot(beta_0)
    Y = Y_hat + rmn(time_cov, space_cov)
    sizes = {"n_C": n_C, "n_T": n_T, "n_V": n_V}
    train = {"beta": beta, "X": X, "Y": Y, "U": U, "X_0": X_0}

    return train, sizes


def test_brsa_rudimentary(seeded_rng):
    """this test is super loose"""

    # this is Mingbo's synth example from the paper
    U = gen_U_nips2016_example()

    n_T = 150
    n_V = 250
    n_nureg = 5

    spacecov_true = np.eye(n_V)

    timecov_true = np.diag(np.abs(norm.rvs(size=(n_T))))

    tr, sz = gen_brsa_data_matnorm_model(
        U,
        n_T=n_T,
        n_V=n_V,
        n_nureg=n_nureg,
        space_cov=spacecov_true,
        time_cov=timecov_true,
    )

    spacecov_model = CovIdentity(size=n_V)
    timecov_model = CovDiagonal(size=n_T)

    model_matnorm = MNRSA(time_cov=timecov_model, space_cov=spacecov_model)

    model_matnorm.fit(tr["Y"], tr["X"], naive_init=False)

    RMSE = np.mean((model_matnorm.C_ - cov2corr(tr["U"])) ** 2) ** 0.5

    assert RMSE < 0.1

    model_matnorm = MNRSA(time_cov=timecov_model, space_cov=spacecov_model)

    model_matnorm.fit(tr["Y"], tr["X"], naive_init=True)

    RMSE = np.mean((model_matnorm.C_ - cov2corr(tr["U"])) ** 2) ** 0.5

    assert RMSE < 0.1



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Generate simulated data for FCMA example

Generate example data for FCMA analyses that uses fmrisim.

This creates two conditions, a and b, with 5 trials for each condtion and
5 participants total. Each trial is 10 TRs long with 7 TRs of stimulation.
Brains are extremely downsampled (10 voxels cubed) in order to help
processing speed.

The signal that discriminates these conditions is such that one region
responds to both conditions whereas another region responds differently to
the two conditions. For instance, imagine voxel X responds to common
activation but voxel A only responds in condition A and voxel B only
responds in condition B.

 Authors: Cameron Ellis (Princeton) 2017
"""
import logging
import numpy as np
from brainiak.utils import fmrisim as sim
import nibabel
import os

logger = logging.getLogger(__name__)


# Default experimental parameters (These can all be changed (will affect
# processing time)
participants = 5  # How many participants are being created
epochs = 5  # How many trials
dimensions = np.array([10, 10, 10])  # What is the size of the brain
stim_dur = 7  # How long is each stimulation period
rest_dur = 3  # How long is the rest between stimulation
conds = 2  # How many conditions are there
fcma_better = 1  # this data is made so fcma will succeed and mvpa will fail

# Where will the data be stored?
directory = 'simulated/'

# Make the directory if it hasn't been made yet
if os.path.isdir(directory) is False:
    os.mkdir(directory)

# Prepare the feature attributes
feature_size = [1]
feature_type = ['cube']
coordinates=[]
coordinates += [np.array(
    [[3, 3, 3], [3, 5, 4]])]
coordinates += [np.array(
    [[3, 3, 3], [6, 5, 4]])]
coordinates += [np.array(
    [[3, 3, 3], [5, 6, 4]])]
signal_magnitude = [1]  # How big is the signal (in SD)


# Inputs for generate_stimfunction
onsets = list(range(conds))
weights = list(range(conds))
tr_duration = 2
event_durations = [tr_duration]
trial_dur = stim_dur + rest_dur
duration = epochs * trial_dur * conds

# Create the epoch cube
epoch = np.zeros([conds, epochs * conds, int(duration / tr_duration)], np.int8,
                 order='C')

# Iterate through the epochs and conditions
for cond_counter in list(range(conds)):

    onsets[cond_counter] = []  # Add a list for this condition
    weights[cond_counter] = []
    for idx in list(range(0, epochs)):

        # When does each epoch start and end
        start_idx = (idx * trial_dur * conds) + (trial_dur * cond_counter)
        end_idx = start_idx + stim_dur

        # Store these start and end times
        onsets[cond_counter] += list(range(start_idx, end_idx, tr_duration))
        epoch[cond_counter, idx * conds + cond_counter, start_idx:end_idx] = 1

        # The pattern of activity for each trial
        weight = ([1] * int(np.floor(stim_dur / 2))) + ([-1]*int(np.ceil(
            stim_dur / 2)))
        weights[cond_counter] += weight

# Iterate through the conditions to make the necessary functions
for cond in list(range(conds)):

    # Generate a volume representing the location and quality of the signal
    volume_signal = sim.generate_signal(dimensions=dimensions,
                                        feature_coordinates=coordinates[cond],
                                        feature_type=feature_type,
                                        feature_size=feature_size,
                                        signal_magnitude=signal_magnitude,
                                        )

    # Create the time course for the signal to be generated
    stimfunction_cond = sim.generate_stimfunction(onsets=onsets[cond],
                                                  event_durations=
                                                  event_durations,
                                                  total_time=duration,
                                                  weights=weights[cond],
                                                  )

    # Convolve the HRF with the stimulus sequence
    signal_function = sim.double_gamma_hrf(stimfunction=stimfunction_cond,
                                           tr_duration=tr_duration,
                                           )

    # Multiply the HRF timecourse with the signal
    signal_cond = sim.apply_signal(signal_function=signal_function,
                                   volume_signal=volume_signal,
                                   )

    # Concatenate all the signal and function files
    if cond == 0:
        stimfunction = stimfunction_cond
        signal = signal_cond
    else:
        stimfunction = list(np.add(stimfunction, stimfunction_cond))
        signal += signal_cond

# Generate the mask of the signal
mask, template = sim.mask_brain(signal)

# Mask the signal to the shape of a brain (does not attenuate signal according
# to grey matter likelihood)
signal *= mask.reshape(dimensions[0], dimensions[1], dimensions[2], 1)

# Downsample the stimulus function to generate it in TR time
stimfunction_tr = stimfunction[::int(tr_duration * 1000)]

# Iterate through the participants and store participants
epochs = []
for participantcounter in range(1, participants + 1):

    # Add the epoch cube
    epochs += [epoch]

    # Save a file name
    savename = directory + 'p' + str(participantcounter) + '.nii'

    # Create the noise volumes (using the default parameters
    noise = sim.generate_noise(dimensions=dimensions,
                               stimfunction_tr=stimfunction_tr,
                               tr_duration=tr_duration,
                               template=template,
                               mask=mask,
                               )

    # Combine the signal and the noise
    brain = signal + noise

    # Save the volume
    affine_matrix = np.diag([-1, 1, 1, 1])  # LR gets flipped
    brain_nifti = nibabel.Nifti1Image(brain, affine_matrix)
    nibabel.save(brain_nifti, savename)

# Save the epochs
np.save(directory + 'epoch_labels.npy', epochs)

# Store the mask
brain_nifti = nibabel.Nifti1Image(mask, affine_matrix)
nibabel.save(brain_nifti, directory + 'mask.nii')



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
import sys
import logging
import scipy.io
import numpy as np
import nibabel as nib
from subprocess import call
from scipy.stats import stats
from nilearn.input_data import NiftiMasker

logging.basicConfig(stream=sys.stdout, level=logging.INFO)

#download data
data_dir = os.path.join(os.getcwd(), 'data')
file_name = os.path.join(data_dir, 's0.mat')
url = ' https://www.dropbox.com/s/r5s9tg4ekxzbrco/s0.mat?dl=0'
cmd = 'curl --location --create-dirs -o ' + file_name + url
try:
    retcode = call(cmd, shell=True)
    if retcode < 0:
        print("File download was terminated by signal", -retcode, file=sys.stderr)
    else:
        print("File download returned", retcode, file=sys.stderr)
except OSError as e:
    print("File download failed:", e, file=sys.stderr)

#get fMRI data and scanner RAS coordinates
all_data = scipy.io.loadmat(file_name)
data = all_data['data']
R = all_data['R']

# Z-score the data
data = stats.zscore(data, axis=1, ddof=1)
n_voxel, n_tr = data.shape

# Run TFA with downloaded data
from brainiak.factoranalysis.tfa import TFA
# uncomment below line to get help message on TFA
#help(TFA)

tfa = TFA(K=5,
        max_num_voxel=int(n_voxel*0.5),
        max_num_tr=int(n_tr*0.5),
        verbose=True)
tfa.fit(data, R)

print("\n centers of latent factors are:")
print(tfa.get_centers(tfa.local_posterior_))
print("\n widths of latent factors are:")
widths = tfa.get_widths(tfa.local_posterior_)
print(widths)
print("\n stds of latent RBF factors are:")
rbf_std = np.sqrt(widths/(2.0))
print(rbf_std)




#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Full Correlation Matrix Analysis (FCMA)

Activity-based voxel selection
"""

# Authors: Yida Wang
# (Intel Labs), 2017

import numpy as np
from sklearn import model_selection
import logging
from mpi4py import MPI

logger = logging.getLogger(__name__)

__all__ = [
    "MVPAVoxelSelector",
]


def _sfn(data, mask, myrad, bcast_var):
    """Score classifier on searchlight data using cross-validation.

    The classifier is in `bcast_var[2]`. The labels are in `bast_var[0]`. The
    number of cross-validation folds is in `bast_var[1].
    """
    clf = bcast_var[2]
    masked_data = data[0][mask, :].T
    # print(l[0].shape, mask.shape, data.shape)
    skf = model_selection.StratifiedKFold(n_splits=bcast_var[1],
                                          shuffle=False)
    accuracy = np.mean(model_selection.cross_val_score(clf, masked_data,
                                                       y=bcast_var[0],
                                                       cv=skf,
                                                       n_jobs=1))
    return accuracy


class MVPAVoxelSelector:
    """Activity-based voxel selection component of FCMA

    Parameters
    ----------

    data: 4D array in shape [brain 3D + epoch]
        contains the averaged and normalized brain data epoch by epoch.
        It is generated by .io.prepare_searchlight_mvpa_data

    mask: 3D array

    labels: 1D array
        contains the labels of the epochs.
        It is generated by .io.prepare_searchlight_mvpa_data

    num_folds: int
        the number of folds to be conducted in the cross validation

    sl: Searchlight
        the distributed Searchlight object
    """
    def __init__(self,
                 data,
                 mask,
                 labels,
                 num_folds,
                 sl
                 ):
        self.data = data
        self.mask = mask.astype(bool)
        self.labels = labels
        self.num_folds = num_folds
        self.sl = sl
        num_voxels = np.sum(self.mask)
        if num_voxels == 0:
            raise ValueError('Zero processed voxels')

    def run(self, clf):
        """ run activity-based voxel selection

        Sort the voxels based on the cross-validation accuracy
        of their activity vectors within the searchlight

        Parameters
        ----------
        clf: classification function
            the classifier to be used in cross validation

        Returns
        -------
        result_volume: 3D array of accuracy numbers
            contains the voxelwise accuracy numbers obtained via Searchlight
        results: list of tuple (voxel_id, accuracy)
            the accuracy numbers of all voxels, in accuracy descending order
            the length of array equals the number of voxels
        """
        rank = MPI.COMM_WORLD.Get_rank()
        if rank == 0:
            logger.info(
                'running activity-based voxel selection via Searchlight'
            )
        self.sl.distribute([self.data], self.mask)
        self.sl.broadcast((self.labels, self.num_folds, clf))
        if rank == 0:
            logger.info(
                'data preparation done'
            )

        # obtain a 3D array with accuracy numbers
        result_volume = self.sl.run_searchlight(_sfn)
        # get result tuple list from the volume
        result_list = result_volume[self.mask]
        results = []
        if rank == 0:
            for idx, value in enumerate(result_list):
                if value is None:
                    value = 0
                results.append((idx, value))
            # Sort the voxels
            results.sort(key=lambda tup: tup[1], reverse=True)
            logger.info(
                'activity-based voxel selection via Searchlight is done'
            )
        return result_volume, results



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""fmrisim real-time generator test script

 Authors: Cameron Ellis (Princeton) 2020
"""
import numpy as np
from brainiak.utils import fmrisim_real_time_generator as gen
from importlib.resources import files
import pytest
import os
import time
import glob

from typing import Dict
from nibabel.nifti1 import Nifti1Image
import gzip

# Test that it crashes without inputs
with pytest.raises(TypeError):
    gen.generate_data()  # type: ignore

data_dict: Dict = {}

rf = files('brainiak').joinpath('utils/sim_parameters/ROI_A.nii.gz')
with rf.open("rb") as f:
    vol = f.read()
data_dict["ROI_A_file"] = np.asanyarray(
    Nifti1Image.from_bytes(gzip.decompress(vol)).dataobj
)

rf = files('brainiak').joinpath('utils/sim_parameters/ROI_B.nii.gz')
with rf.open("rb") as f:
    vol = f.read()
data_dict["ROI_B_file"] = np.asanyarray(
    Nifti1Image.from_bytes(gzip.decompress(vol)).dataobj
)
rf = files('brainiak').joinpath('utils/sim_parameters/sub_template.nii.gz')
with rf.open("rb") as f:
    vol = f.read()
data_dict["template_path"] = np.asanyarray(
    Nifti1Image.from_bytes(gzip.decompress(vol)).dataobj
)

rf = files('brainiak').joinpath('utils/sim_parameters/sub_noise_dict.txt')
with rf.open("rb") as f:
    noise_dict_file = f.read()

data_dict['noise_dict_file'] = noise_dict_file
data_dict['numTRs'] = 30
data_dict['event_duration'] = 2
data_dict['scale_percentage'] = 1
data_dict['different_ROIs'] = True
data_dict['multivariate_pattern'] = False
data_dict['save_dicom'] = False
data_dict['save_realtime'] = False
data_dict['trDuration'] = 2
data_dict['isi'] = 4
data_dict['burn_in'] = 6


# Run default test
def test_default(tmp_path, dd=data_dict):

    # copy data_dict so values aren't changed
    dd = dd.copy()

    # Run the simulation
    gen.generate_data(str(tmp_path),
                      dd)

    # Check that there are 32 files where there should be (30 plus label and
    # mask)
    assert len(os.listdir(str(tmp_path))) == 32, "Incorrect file number"

    # Check that the data is the right shape
    input_template = dd['template_path']
    input_shape = input_template.shape
    output_vol = np.load(tmp_path / 'rt_000.npy')
    output_shape = output_vol.shape
    assert input_shape == output_shape, 'Output shape is incorrect'

    # Check the labels have the correct count
    labels = np.load(tmp_path / 'labels.npy')

    assert np.sum(labels > 0) == 9, 'Incorrect number of events'


def test_signal_size(tmp_path, dd=data_dict):

    dd = dd.copy()

    # Change it to only use ROI A
    dd['different_ROIs'] = False

    # Make the signal large
    dd['scale_percentage'] = 100

    # Run the simulation
    gen.generate_data(str(tmp_path),
                      dd)

    # Load in the ROI masks
    ROI_A = dd['ROI_A_file']
    ROI_B = dd['ROI_B_file']

    # Load in the data just simulated
    ROI_A_mean = []
    ROI_B_mean = []
    for TR_counter in range(dd['numTRs']):

        # Load the data
        vol_name = 'rt_%03d.npy' % TR_counter
        vol = np.load(tmp_path / vol_name)

        # Mask the data
        ROI_A_mean += [np.mean(vol[ROI_A == 1])]
        ROI_B_mean += [np.mean(vol[ROI_B == 1])]

    assert np.std(ROI_A_mean) > np.std(ROI_B_mean), 'Signal not scaling'


def test_multivariate(tmp_path, dd=data_dict):

    dd = dd.copy()

    dd['multivariate_pattern'] = True
    dd['different_ROIs'] = False

    # Make the signal large
    dd['scale_percentage'] = 100

    # Run the simulation
    gen.generate_data(str(tmp_path),
                      dd)

    # Load in the ROI masks
    ROI_A = dd['ROI_A_file']
    ROI_B = dd['ROI_B_file']

    # Test this volume
    vol = np.load(str(tmp_path / 'rt_007.npy'))

    ROI_A_std = np.std(vol[ROI_A == 1])
    ROI_B_std = np.std(vol[ROI_B == 1])

    assert ROI_A_std > ROI_B_std, 'Multivariate not making variable signal'


def test_save_dicoms_realtime(tmp_path, dd=data_dict):

    dd = dd.copy()
    start_time = time.time()

    dd['save_dicom'] = True
    dd['save_realtime'] = True

    # test when ROI files are not set
    dd['ROI_A_file'] = None
    dd['ROI_B_file'] = None
    dd['template_path'] = None
    dd['noise_dict_file'] = None

    # Run the simulation
    gen.generate_data(str(tmp_path),
                      dd)

    end_time = time.time()

    # Check it took 2s per TR
    assert (end_time - start_time) > 60, 'Realtime ran fast'

    # Check correct file number
    file_path = str(tmp_path / '*.dcm')
    assert len(glob.glob(file_path)) == 30, "Wrong dicom file num"



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from brainiak.fcma.mvpa_voxelselector import MVPAVoxelSelector
from brainiak.fcma.preprocessing import prepare_searchlight_mvpa_data
from brainiak import io
from sklearn import svm
import sys
from mpi4py import MPI
import logging
import nibabel as nib
import numpy as np
from brainiak.searchlight.searchlight import Searchlight

format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
# if want to output log to a file instead of outputting log to the console,
# replace "stream=sys.stdout" with "filename='fcma.log'"
logging.basicConfig(level=logging.INFO, format=format, stream=sys.stdout)
logger = logging.getLogger(__name__)

"""
example running command in run_mvpa_voxel_selection.sh
"""
if __name__ == '__main__':
    if MPI.COMM_WORLD.Get_rank()==0:
        logger.info(
            'programming starts in %d process(es)' %
            MPI.COMM_WORLD.Get_size()
        )
    if len(sys.argv) != 6:
        logger.error('the number of input argument is not correct')
        sys.exit(1)

    data_dir = sys.argv[1]
    suffix = sys.argv[2]
    mask_file = sys.argv[3]
    epoch_file = sys.argv[4]

    # all MPI processes read the mask; the mask file is small
    mask_image = nib.load(mask_file)
    mask = io.load_boolean_mask(mask_file)
    data = None
    labels = None
    if MPI.COMM_WORLD.Get_rank()==0:
        logger.info(
            'mask size: %d' %
            np.sum(mask)
        )
        images = io.load_images_from_dir(data_dir, suffix=suffix)
        conditions = io.load_labels(epoch_file)
        data, labels = prepare_searchlight_mvpa_data(images, conditions)

        # setting the random argument produces random voxel selection results
        # for non-parametric statistical analysis.
        # There are three random options:
        # RandomType.NORANDOM is the default
        # RandomType.REPRODUCIBLE permutes the voxels in the same way every run
        # RandomType.UNREPRODUCIBLE permutes the voxels differently across runs
        # example
        #from brainiak.fcma.preprocessing import RandomType
        #data, labels = prepare_searchlight_mvpa_data(images, conditions,
        #                                                    random=RandomType.UNREPRODUCIBLE)

        # the following line is an example to leaving a subject out
        #epoch_info = [x for x in epoch_info if x[1] != 0]

    num_subjs = int(sys.argv[5])
    # create a Searchlight object
    sl = Searchlight(sl_rad=1)
    mvs = MVPAVoxelSelector(data, mask, labels, num_subjs, sl)
    clf = svm.SVC(kernel='linear', shrinking=False, C=1, gamma='auto')
    # only rank 0 has meaningful return values
    score_volume, results = mvs.run(clf)
    # this output is just for result checking
    if MPI.COMM_WORLD.Get_rank()==0:
        score_volume = np.nan_to_num(score_volume.astype(np.float))
        io.save_as_nifti_file(score_volume, mask_image.affine,
                                   'result_score.nii.gz')
        seq_volume = np.zeros(mask.shape, dtype=int)
        seq = np.zeros(len(results), dtype=int)
        with open('result_list.txt', 'w') as fp:
            for idx, tuple in enumerate(results):
                fp.write(str(tuple[0]) + ' ' + str(tuple[1]) + '\n')
                seq[tuple[0]] = idx
        seq_volume[mask] = seq
        io.save_as_nifti_file(seq_volume, mask_image.affine,
                                   'result_seq.nii.gz')



#  Copyright 2018 David Huberdeau & Peter Kok
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Copyright 2020 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

# Authors: David Huberdeau (Yale University) &
# Peter Kok (Yale University), 2018 &
# Vy Vo (Intel Corp., UCSD), 2019

import pytest
import numpy as np
import logging
from brainiak.reconstruct.iem import InvertedEncoding1D, InvertedEncoding2D
from brainiak.utils.fmrisim import generate_1d_gaussian_rfs, \
    generate_1d_rf_responses
from scipy.stats import circmean

logger = logging.getLogger(__name__)


# Simple test: can an instance be instantiated?
def test_can_instantiate():
    s = InvertedEncoding1D()
    assert s, "Invalid InvertedEncoding1D instance"
    s2 = InvertedEncoding2D(stim_xlim=[0, 1], stim_ylim=[0, 1],
                            stimulus_resolution=[1, 1])
    assert s2, "Invalid InvertedEncoding2D instance"


# Test for checking range values.
def test_instantiate_improper_range():
    with pytest.raises(ValueError):
        s = InvertedEncoding1D(6, 5, 'halfcircular', range_start=20,
                               range_stop=0)
        assert s, "Invalid InvertedEncoding1D instance"
    with pytest.raises(ValueError):
        s2 = InvertedEncoding2D(stim_xlim=[0, -1], stim_ylim=[0, -1],
                                stimulus_resolution=[10, 10])
        assert s2, "Invalid InvertedEncoding2D instance"
    with pytest.raises(ValueError):
        s2 = InvertedEncoding2D(stim_xlim=[0], stim_ylim=[-1, 0],
                                stimulus_resolution=10)
        assert s2, "Invalid InvertedEncoding2D instance"


# Test for n_observations < n_channels
def test_data_amount():
    x = np.random.rand(5, 1000)
    s = InvertedEncoding1D()
    with pytest.raises(ValueError):
        s.fit(x, np.random.rand(5))
        assert s, "Invalid data"
    s2 = InvertedEncoding2D(stim_xlim=[-1, 1], stim_ylim=[-1, 1],
                            stimulus_resolution=10)
    with pytest.raises(ValueError):
        s2.fit(x, np.random.rand(5))


# Test number of data dimensions
def test_data_dimensions():
    x = np.random.rand(5, 10, 2)
    s = InvertedEncoding1D()
    with pytest.raises(ValueError):
        s.fit(x, np.random.rand(5))
    s2 = InvertedEncoding2D(stim_xlim=[-1, 1], stim_ylim=[-1, 1],
                            stimulus_resolution=10)
    with pytest.raises(ValueError):
        s2.fit(x, np.random.rand(5))


# TESTS FOR 2D MODEL #
# Test to check that stimulus resolution is used properly
def test_2d_stimulus_resolution():
    s2 = InvertedEncoding2D(stim_xlim=[-1, 1], stim_ylim=[-1, 1],
                            stimulus_resolution=10)
    assert len(s2.stim_pixels[0] == 10)
    assert len(s2.stim_pixels[1] == 10)
    s2 = InvertedEncoding2D(stim_xlim=[-1, 1], stim_ylim=[-2, 2],
                            stimulus_resolution=[10, 20])
    assert len(s2.stim_pixels[0] == 10)
    assert len(s2.stim_pixels[1] == 20)


# Test that 2D channels can be set by the user
def test_2d_custom_channels():
    nchan = 8
    res = 10
    npix = res*res
    channels = np.random.rand(nchan, npix)*2 - 1
    bds = [-1, 1]
    s = InvertedEncoding2D(stim_xlim=bds, stim_ylim=bds,
                           stimulus_resolution=res, chan_xlim=bds,
                           chan_ylim=bds, channels=channels)
    assert s, "Unable to define custom InvertedEncoding2D channels"


# Test that channel definition should be consistent.
def test_cannot_instantiate_2d_channels():
    # Channel definition over wrong number of pixels (5 instead of 100)
    with pytest.raises(ValueError):
        s = InvertedEncoding2D(stim_xlim=[-1, 1], stim_ylim=[-1, 1],
                               stimulus_resolution=10,
                               channels=np.random.rand(5, 5))
        assert s, "Invalid InvertedEncoding2D instance"


# Test that you cannot modify properties in an inconsistent way.
def test_modify_2d_properties():
    nchan = 8
    res = 10
    npix = res*res
    channels = np.random.rand(nchan, npix)*2 - 1
    bds = [-1, 1]
    s = InvertedEncoding2D(stim_xlim=bds, stim_ylim=bds,
                           stimulus_resolution=res, chan_xlim=bds,
                           chan_ylim=bds, channels=channels)
    with pytest.raises(ValueError):
        s = s.set_params(n_channels=nchan - 1)
        assert s, "Invalid InvertedEncoding2D instance"
    with pytest.raises(ValueError):
        s = s.set_params(xp=np.random.rand(npix - 10))
        assert s, "Invalid InvertedEncoding2D instance"
    with pytest.raises(ValueError):
        s = s.set_params(stim_fov=[[0, 1], [0, -1]])
        assert s, "Invalid InvertedEncoding2D instance"
    with pytest.raises(ValueError):
        s = s.set_params(stim_fov=[[0, 1]])
        assert s, "Invalid InvertedEncoding2D instance"
    with pytest.raises(ValueError):
        s = s.set_params(stim_fov=[[0], [0, 1]])
        assert s, "Invalid InvertedEncoding2D instance"


# Test that you can get object properties
def test_get_2d_params():
    bds = [-1, 1]
    res = 10
    s = InvertedEncoding2D(stim_xlim=bds, stim_ylim=bds,
                           stimulus_resolution=res)
    param_out = s.get_params()
    assert np.all(param_out.get('stim_fov')[0] == bds)
    assert param_out.get('xp').size == res*res


# Test helper function to create 2D cosine
def test_2d_cos():
    nchan = 8
    res = 10
    npix = res*res
    bds = [-1, 1]
    sz = 2
    s = InvertedEncoding2D(stim_xlim=bds, stim_ylim=bds,
                           stimulus_resolution=res,
                           channels=np.random.rand(nchan, npix))
    sz = s._2d_cosine_fwhm_to_sz(1)
    fcn = s._make_2d_cosine(s.xp.reshape(-1, 1), s.yp.reshape(-1, 1),
                            np.linspace(bds[0], bds[1], nchan),
                            np.linspace(bds[0], bds[1], nchan), sz)
    assert fcn.shape == (nchan, npix)
    # Test that masking works -- basis function should have fewer non-zero
    # elements than specified by the size constant
    xd = np.diff(s.xp)[0][0]
    nval = (np.nonzero(fcn[0, :])[0]).size
    assert nval*(xd**2) <= sz**2


# Test size conversion functions
def test_2d_cos_size_fcns():
    bds = [-1, 1]
    s = np.random.rand()
    imodel = InvertedEncoding2D(stim_xlim=bds, stim_ylim=bds,
                                stimulus_resolution=10)
    fwhm = imodel._2d_cosine_sz_to_fwhm(s)
    s2 = imodel._2d_cosine_fwhm_to_sz(fwhm)
    assert np.isclose(s, s2)
    fwhm2 = imodel._2d_cosine_sz_to_fwhm(s2)
    assert np.isclose(fwhm, fwhm2)


def test_square_basis_grid():
    nchan = 8
    bds = [-1, 1]
    s = InvertedEncoding2D(stim_xlim=bds, stim_ylim=bds,
                           stimulus_resolution=10)
    _, centers = s.define_basis_functions_sqgrid(nchannels=nchan)
    assert centers.shape[0] == nchan*nchan
    xspacing = np.round(np.diff(centers[:, 0]), 5)
    yspacing = np.round(np.diff(centers[:, 1]), 5)
    assert xspacing[0] == xspacing[28] == xspacing[-1]
    assert yspacing[0] == yspacing[25] == yspacing[-1]


def test_triangular_basis_grid():
    grid_rad = 3
    n_channels = (grid_rad*2 + 1) * (grid_rad*2)
    bds = [-1, 1]
    s = InvertedEncoding2D(stim_xlim=bds, stim_ylim=bds,
                           stimulus_resolution=10)
    _, centers = s.define_basis_functions_trigrid(grid_rad)
    assert centers.shape[0] == n_channels
    xspacing = np.round(np.diff(centers[:, 0]), 4)
    assert xspacing[0] == xspacing[np.random.randint(n_channels)] == \
        xspacing[-1]
    ysp = xspacing[0] * np.sqrt(3) * 0.5
    yspacing = np.diff(centers[:, 1])
    yspace = yspacing[yspacing > 0.0]
    assert np.all((ysp - yspace) < 1e-5)


# Define some data to use in the following tests.
nobs, nvox, ntest = 100, 1000, 5
xlim, ylim = [[-6, 6], [-3, 3]]
res = [100, 100]
sxx, syy = np.meshgrid(np.linspace(xlim[0], xlim[1], 10),
                       np.linspace(ylim[0], ylim[1], 10))
yd = np.hstack((sxx.reshape(-1, 1), syy.reshape(-1, 1)))
Xd = np.zeros((nobs, nvox))
for i, l in enumerate(np.linspace(-1, 1, 10)):
    Xd[i*10:i*10+10, :] = np.random.normal(loc=l, scale=1.5,
                                           size=(10, nvox))
X2d = np.zeros((ntest, nvox))
for i, l in enumerate(np.linspace(-1, 1, 5)):
    X2d[i, :] = np.random.normal(loc=l, scale=1.5,
                                 size=(1, nvox))


# Test that 2D model raises error if design matrix C cannot be defined
def test_fit_invalid_2d():
    # C=None and stim_radius=None here, cannot define C
    i2 = InvertedEncoding2D(stim_xlim=xlim, stim_ylim=ylim,
                            stimulus_resolution=res, stim_radius=None)
    i2.define_basis_functions_sqgrid(nchannels=[12, 6])
    with pytest.raises(ValueError):
        i2.fit(Xd, yd)


# Test attempt to fit with list of varying stimulus radii
def test_fit_2d_radius_list():
    i2 = InvertedEncoding2D(stim_xlim=xlim, stim_ylim=ylim,
                            stimulus_resolution=res,
                            stim_radius=np.random.rand(nobs))
    i2.define_basis_functions_sqgrid(nchannels=[12, 6])
    i2.fit(Xd, yd)


# Test with custom C input
def test_fit_custom_channel_activations():
    i2 = InvertedEncoding2D(stim_xlim=xlim, stim_ylim=ylim,
                            stimulus_resolution=res,
                            stim_radius=12)
    i2.define_basis_functions_sqgrid(nchannels=[12, 6])
    # Define C by expanding y & adding noise to avoid singular W matrix error
    C0 = np.repeat(np.expand_dims(yd[:, 0], 1), 12*3, axis=1) + \
        np.random.rand(nobs, 12*3)
    C1 = np.repeat(np.expand_dims(yd[:, 1], 1), 12*3, axis=1) + \
        np.random.rand(nobs, 12*3)
    i2.fit(Xd, yd, np.hstack((C0, C1)))
    assert np.all(i2.W_)


iem_2d = InvertedEncoding2D(stim_xlim=xlim, stim_ylim=ylim,
                            stimulus_resolution=res, stim_radius=12)
iem_2d.define_basis_functions_sqgrid(nchannels=[12, 6])


# Test if valid data can be fit.
def test_can_fit_2d_data():
    iem_2d.fit(Xd, yd)


# Show that a data matrix with improper format (dimensions) breaks the
# algorithm.
def test_cannot_fit_2d_data():
    with pytest.raises(ValueError):
        iem_2d.fit(Xd.transpose(), yd)


# Ill conditioned data matrix will raise error
def test_ill_conditioned_2d_train_data():
    with pytest.raises(ValueError):
        Xt = np.ones((nobs, nvox))
        y = np.random.rand(nobs, 2)
        iem_2d.fit(Xt, y)


# Ill conditioned channel activations C will raise warning
def test_ill_conditioned_2d_channel_activations():
    with pytest.warns(RuntimeWarning):
        C = iem_2d._define_trial_activations(np.ones((nobs, 2)))
        assert np.linalg.matrix_rank(C) == 1


# Ill conditioned weight matrix will raise error
def test_ill_conditioned_2d_weights():
    with pytest.raises(ValueError):
        Xt = np.random.rand(nobs, nvox)
        y = np.random.rand(nobs, 2)
        iem_2d.fit(Xt, y)


# Not enough observations will trigger error
def test_insufficient_2d_data():
    with pytest.raises(ValueError):
        Xt = np.random.rand(10, nvox)
        y = np.random.rand(10, 2)
        iem_2d.fit(Xt, y)


# Test case when # of observations are not matched btwn data & labels
def test_mismatched_2d_observations():
    with pytest.raises(ValueError):
        iem_2d.fit(Xd, yd[:-50, :])


# Test prediction capability from valid (fabricated) data
def test_can_predict_from_2d_data():
    iem_2d.fit(Xd, yd)
    preds = iem_2d.predict(X2d)
    assert preds.shape == (ntest, 2)


# Show that prediction is invalid when input data is wrong size
def test_cannot_predict_from_2d_data():
    iem_2d.fit(Xd, yd)
    with pytest.raises(ValueError):
        _ = iem_2d.predict(X2d.T)


# Show proper scoring function with valid (fabricated) test data
def test_can_score_2d():
    iem_2d.fit(Xd, yd)
    score = iem_2d.score(X2d, yd[:ntest, :])
    assert score.shape == (ntest,)
    score = iem_2d.score_against_reconstructed(X2d,
                                               np.random.rand(res[0]*res[1],
                                                              ntest))
    assert score.shape == (ntest,)
    score = iem_2d.score_against_reconstructed(X2d,
                                               np.random.rand(res[0]*res[1],
                                                              ntest),
                                               metric="cosine")
    assert score.shape == (ntest,)


# Test scoring with invalid data formatting
def test_cannot_score_2d():
    iem_2d.fit(Xd, yd)
    with pytest.raises(ValueError):
        score = iem_2d.score(X2d.transpose(), yd[ntest, :])
        assert score


# TESTS FOR 1D MODEL #
# Test to check stimulus resolution input
def test_1d_stimulus_resolution():
    s = InvertedEncoding1D(6, 5, stimulus_resolution=360)
    assert s.stim_res == 360


# Provide invalid data so that channels cannot be created.
def test_cannot_instantiate_1d_channels():
    with pytest.raises(ValueError):
        s = InvertedEncoding1D(n_channels=0)
        assert s, "Invalid InvertedEncoding1D instance"


# Provide invalid stimulus mode
def test_stimulus_mode():
    with pytest.raises(ValueError):
        s = InvertedEncoding1D(6, 5, 'random')
        assert s, "Invalid InvertedEncoding1D instance"


# Provide mismatching range and stimulus_mode input
def test_range_stimulus_mode_circ():
    with pytest.raises(ValueError):
        s = InvertedEncoding1D(6, 5, 'circular', 0, 180)
        assert s, "Invalid InvertedEncoding1D instance"


# Provide mismatching range & stimulus mode, with half circular
def test_range_stimulus_mode_halfcirc():
    with pytest.raises(ValueError):
        s = InvertedEncoding1D(6, 5, 'halfcircular', -10, 350)
        assert s, "Invalid InvertedEncoding1D instance"


# Define some data to use in the following tests.
n, dim = 297, 9
n_ = n // dim
y = np.repeat(np.linspace(0, 180-(180/dim), dim), n_)
voxel_rfs, _ = generate_1d_gaussian_rfs(dim, 180, (0, 179),
                                        random_tuning=False)
X = generate_1d_rf_responses(voxel_rfs, y, 180, (0, 179),
                             trial_noise=0.25).transpose()
X2 = generate_1d_rf_responses(voxel_rfs, y, 180, (0, 179),
                              trial_noise=0.25).transpose()


# Test if valid data can be fit.
def test_can_fit_data():
    Invt_model = InvertedEncoding1D()
    Invt_model.fit(X, y)


# Test if valid data can be fit in circular space.
def test_can_fit_circular_space():
    s = InvertedEncoding1D(6, 5, 'circular', range_stop=360)
    s.fit(X, y)


# Show that a data matrix with improper format (dimensions) breaks the
# algorithm.
def test_cannot_fit_data():
    with pytest.raises(ValueError):
        Invt_model = InvertedEncoding1D()
        Invt_model.fit(X.transpose(), y)


def test_ill_conditioned_train_data():
    Invt_model = InvertedEncoding1D()
    with pytest.raises(ValueError):
        X = np.array([[0, 0, 0], [1, 1, 1]])
        Invt_model.fit(X, np.array([0, 0, 0]))


# Test case if data dimensions are wrong
def test_extra_data_dimensions():
    with pytest.raises(ValueError):
        n, dim1, dim2 = 300, 3, 3
        X = np.random.rand(n//3, dim1, dim2)
        Invt_model = InvertedEncoding1D()
        Invt_model.fit(X, y)


# Test case when # of observations are not matched btwn data & labels
def test_mismatched_observations():
    with pytest.raises(ValueError):
        Invt_model = InvertedEncoding1D()
        Invt_model.fit(X, y[:-50])


# Test prediction capability from valid (fabricated) data
def test_can_predict_from_data():
    Invt_model = InvertedEncoding1D()
    Invt_model.fit(X, y)
    m_reconstruct = []
    for j in np.arange(dim):
        preds = Invt_model.predict(X2[n_*j:n_*(j+1), :])
        tmp = circmean(np.deg2rad(preds))
        m_reconstruct.append(np.rad2deg(tmp))
    logger.info('Reconstructed angles: ' + str(m_reconstruct))


# Show that prediction is invalid when input data is wrong size
def test_cannot_predict_from_data():
    Invt_model = InvertedEncoding1D()
    Invt_model.fit(X, y)
    with pytest.raises(ValueError):
        _ = Invt_model.predict(X2[0:n_, :].transpose())


# Show proper scoring function with valid (fabricated) test data
def test_can_score():
    Invt_model = InvertedEncoding1D()
    Invt_model.fit(X, y)
    score = Invt_model.score(X2, y)
    logger.info('Scores: ' + str(score))


# Test scoring with invalid data formatting
def test_cannot_score():
    with pytest.raises(ValueError):
        Invt_model = InvertedEncoding1D()
        Invt_model.fit(X, y)
        score = Invt_model.score(X2.transpose(), y)
        logger.info('Scores: ' + str(score))


# Test stimulus resolution that is not even multiple
def test_stimulus_resolution_odd():
    Invt_model = InvertedEncoding1D(stimulus_resolution=59)
    with pytest.raises(NotImplementedError):
        Invt_model.fit(X, y)


# Test stimulus masking
def test_stimulus_mask():
    Invt_model = InvertedEncoding1D(6, 5, range_start=-10,
                                    range_stop=170,
                                    stimulus_resolution=60)
    chans, _ = Invt_model._define_channels()
    Invt_model.set_params(channels_=chans)
    with pytest.warns(RuntimeWarning):
        C = Invt_model._define_trial_activations(np.array([50]))
        tmp_C = np.repeat([0, 1, 0], 60) @ chans.transpose()
        assert np.all((C - tmp_C) < 1e-7)


# Test stimulus masking with different range
def test_stimulus_mask_shift_positive():
    Invt_model = InvertedEncoding1D(6, 5, range_start=10,
                                    range_stop=190,
                                    stimulus_resolution=60)
    chans, _ = Invt_model._define_channels()
    Invt_model.set_params(channels_=chans)
    with pytest.warns(RuntimeWarning):
        C = Invt_model._define_trial_activations(np.array([70]))
        tmp_C = np.repeat([0, 1, 0], 60) @ chans.transpose()
        assert np.all((C - tmp_C) < 1e-7)


# Test ability to get model parameters from object
def test_can_get_params():
    s = InvertedEncoding1D()
    param_out = s.get_params()
    assert param_out.get('channel_exp') == 5
    logger.info('Returned Parameters: ' +
                str(param_out.get('n_channels')) +
                ', ' + str(param_out.get('range_start')) +
                ', ' + str(param_out.get('range_stop')))


# Test ability to set model parameters of an object instance
def test_can_set_params():
    s = InvertedEncoding1D()
    s.set_params(n_channels=10,
                 stimulus_mode='circular',
                 range_start=-90,
                 range_stop=270,
                 channel_exp=4,
                 verbose=False)



# The following code is designed to perform a searchlight at every voxel in the brain looking at the difference in pattern similarity between musical genres (i.e. classical and jazz). In the study where the data was obtained, subjects were required to listen to a set of 16 songs twice (two runs) in an fMRI scanner. The 16 songs consisted of 8 jazz songs and 8 classical songs. The goal of this searchlight is to find voxels that seem to represent distinct information about these different musical genres. Presumably, these voxels would be found in the auditory cortex which happens to be the most organized system in the brain for processing sound information.

import numpy as np
import time
from mpi4py import MPI
from nilearn.image import load_img
import sys
from brainiak.searchlight.searchlight import Searchlight
from scipy import stats
from scipy.sparse import random
import os

# MPI variables
comm = MPI.COMM_WORLD
rank = comm.rank
size = comm.size

# Generate random data
if rank == 0:
    np.random.seed(0)
    data1_rand = np.random.rand(91,109,91,16)
    data2_rand = np.random.rand(91,109,91,16)
    classical = np.random.rand(2600)
    jazz = np.random.rand(2600)
    d1_reshape = np.reshape(data1_rand,(91*109*91,16))
    d2_reshape = np.reshape(data2_rand,(91*109*91,16))
    a1 = load_img('a1plus_2mm.nii.gz')
    a1_vec = np.reshape(a1.get_fdata(),(91*109*91))
    a1_idx = np.nonzero(a1_vec)
    for i in range(8):
        d1_reshape[a1_idx[0],i] += classical
        d1_reshape[a1_idx[0],i+8] += jazz
        d2_reshape[a1_idx[0],i] += classical
        d2_reshape[a1_idx[0],i+8] += jazz
    data1 = np.reshape(d1_reshape,(91,109,91,16))
    data2 = np.reshape(d2_reshape,(91,109,91,16))

    # Flatten data, then zscore data, then reshape data back into MNI coordinate space
    data1 = stats.zscore(np.reshape(data1,(91*109*91,16)))
    data1 = np.reshape(data1,(91,109,91,16))
    data2 = stats.zscore(np.reshape(data2,(91*109*91,16)))
    data2 = np.reshape(data2,(91,109,91,16))
else:
    data1 = None
    data2 = None

# Load mask
mask_img = load_img('MNI152_T1_2mm_brain_mask.nii')
mask_img = mask_img.get_fdata()

# Definte function that takes the difference between within vs. between genre comparisons
def corr2_coeff(AB,msk,myrad,bcast_var):
    if not np.all(msk):
        return None
    A,B = (AB[0], AB[1])
    A = A.reshape((-1,A.shape[-1]))
    B = B.reshape((-1,B.shape[-1]))
    corrAB = np.corrcoef(A.T,B.T)[16:,:16]
    classical_within = np.mean(corrAB[0:8,0:8])
    jazz_within = np.mean(corrAB[8:16,8:16])
    classJazz_between = np.mean(corrAB[8:16,0:8])
    jazzClass_between = np.mean(corrAB[0:8,8:16])
    within_genre = np.mean([classical_within,jazz_within])
    between_genre = np.mean([classJazz_between,jazzClass_between])
    diff = within_genre - between_genre
    return diff

comm.Barrier()
begin_time = time.time()
comm.Barrier()

# Create and run searchlight
sl = Searchlight(sl_rad=1,max_blk_edge=5)
sl.distribute([data1,data2],mask_img)
sl.broadcast(None)
global_outputs = sl.run_searchlight(corr2_coeff)

comm.Barrier()
end_time = time.time()
comm.Barrier()

# Plot searchlight results
if rank == 0:
    print('Searchlight Done: ', end_time - begin_time)
    maxval = np.max(global_outputs[np.not_equal(global_outputs,None)])
    minval = np.min(global_outputs[np.not_equal(global_outputs,None)])
    global_outputs = np.array(global_outputs, dtype=np.float)
    print(global_outputs)

    # Save searchlight images
    out_dir = "searchlight_images"
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    import matplotlib.pyplot as plt
    for (cnt, img) in enumerate(global_outputs):
        plt.imshow(img,vmin=minval,vmax=maxval)
        plt.colorbar()
        plt.savefig('searchlight_images/' + 'img' + str(cnt) + '.png')
        plt.clf()





import multiprocessing
import sys

from mpi4py import MPI

import pytest
import numpy
import random
import tensorflow


pytest_plugins = ["tests.pytest_mpiexec_plugin"]


def pytest_configure(config):
    config.option.xmlpath = "junit-{}.xml".format(MPI.COMM_WORLD.Get_rank())


def pytest_addoption(parser):
    parser.addoption(
        "--enable_notebook_tests",
        action="store_true",
        dest="enable_notebook_tests",
        default=False,
        help="Enable tests for Jupyter notebook examples in docs/examples. "
        "These can take a long time and are disabled by default.",
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--enable_notebook_tests"):
        # --enable_notebook_tests given in cli: do not skip notebook tests
        return
    else:
        enable_notebook_tests = pytest.mark.skip(
            reason="needs --enable_notebook_tests option to run"
        )
        for item in items:
            if "notebook" in item.keywords:
                item.add_marker(enable_notebook_tests)


@pytest.fixture
def seeded_rng():
    random.seed(0)
    numpy.random.seed(0)
    tensorflow.random.set_seed(0)


@pytest.fixture(scope="module", autouse=True)
def pool_size():
    """
    Set the pool_size to 1 for MPI tests when start_method for multiprocessing
    is not fork.

    This replaces the old skip_non_fork fixture. We don't need to skip these
    tests completely, but we need to ensure that the pool_size is set to 1 so
    they don't launch any multiprocessing pools within the MPI environment.
    On windows, it seems like intel mpi and msmpi both have issues with fork,
    so we need to set the pool_size to 1 there as well.
    """
    if (multiprocessing.get_start_method() != "fork" and
            MPI.COMM_WORLD.Get_attr(MPI.APPNUM) is not None):
        return 1

    # OpenMPI has issues with fork, so we need to set the pool_size to 1
    if "Open MPI" in MPI.get_vendor()[0]:
        return 1

    # On Windows, we need to set the pool_size to 1 for intel mpi and msmpi
    elif sys.platform == "win32":
        return 1

    else:
        return 2



"""-----------------------------------------------------------------------------

initialize.py (Last Updated: 10/16/2020)

The purpose of this script is to initialize the rt-cloud session. Specifically,
it will initiate any variables that need to be initiated (e.g., configuration
file) and upload any necessary files to the cloud.

-----------------------------------------------------------------------------"""

# print a short introduction on the internet window
print(""
    "-----------------------------------------------------------------------------\n"
    "The initialize script has not been implemented for this project.\n\n"
    "To implement, please edit initialize.py. For an example as to how you can\n"
    "do this, refer to the initialize.py script in the sample project for the\n"
    "rtcloud framework. Go to https://github.com/brainiak/rt-cloud.git for the\n"
    "repo and navigate to projects/sample/\n"
    "-----------------------------------------------------------------------------")


#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from sklearn import svm
#from sklearn.linear_model import LogisticRegression
import sys
import logging
from brainiak.fcma.preprocessing import prepare_mvpa_data
from brainiak import io
import numpy as np
from scipy.spatial.distance import hamming
from sklearn import model_selection
#from sklearn.externals import joblib

format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
# if want to output log to a file instead of outputting log to the console,
# replace "stream=sys.stdout" with "filename='fcma.log'"
logging.basicConfig(level=logging.INFO, format=format, stream=sys.stdout)
logger = logging.getLogger(__name__)

# python3 mvpa_classification.py face_scene bet.nii.gz face_scene/visual_top_mask.nii.gz face_scene/fs_epoch_labels.npy
if __name__ == '__main__':
    if len(sys.argv) != 5:
        logger.error('the number of input argument is not correct')
        sys.exit(1)

    data_dir = sys.argv[1]
    extension = sys.argv[2]
    mask_file = sys.argv[3]
    epoch_file = sys.argv[4]

    epoch_list = np.load(epoch_file)
    num_subjects = len(epoch_list)
    num_epochs_per_subj = epoch_list[0].shape[1]

    logger.info(
        'doing MVPA training and classification on %d subjects, each of which has %d epochs' %
        (num_subjects, num_epochs_per_subj)
    )

    images = io.load_images_from_dir(data_dir, extension)
    mask = io.load_boolean_mask(mask_file)
    conditions = io.load_labels(epoch_file)
    processed_data, labels = prepare_mvpa_data(images, conditions, mask)

    # transpose data to facilitate training and prediction
    processed_data = processed_data.T

    clf = svm.SVC(kernel='linear', shrinking=False, C=1, gamma='auto')
    # doing leave-one-subject-out cross validation
    for i in range(num_subjects):
        leave_start = i * num_epochs_per_subj
        leave_end = (i+1) * num_epochs_per_subj
        training_data = np.concatenate((processed_data[0:leave_start], processed_data[leave_end:]), axis=0)
        test_data = processed_data[leave_start:leave_end]
        training_labels = np.concatenate((labels[0:leave_start], labels[leave_end:]), axis=0)
        test_labels = labels[leave_start:leave_end]
        clf.fit(training_data, training_labels)
        # joblib can be used for saving and loading models
        #joblib.dump(clf, 'model/logistic.pkl')
        #clf = joblib.load('model/svm.pkl')
        predict = clf.predict(test_data)
        print(predict)
        print(clf.decision_function(test_data))
        print(np.asanyarray(test_labels))
        incorrect_predict = hamming(predict, np.asanyarray(test_labels)) * num_epochs_per_subj
        logger.info(
            'when leaving subject %d out for testing, the accuracy is %d / %d = %.2f' %
            (i, num_epochs_per_subj-incorrect_predict, num_epochs_per_subj,
             (num_epochs_per_subj-incorrect_predict) * 1.0 / num_epochs_per_subj)
        )

    # use model selection
    # no shuffling in cv
    skf = model_selection.StratifiedKFold(n_splits=num_subjects,
                                          shuffle=False)
    scores = model_selection.cross_val_score(clf, processed_data,
                                             y=labels,
                                             cv=skf)
    print(scores)
    logger.info(
        'the overall cross validation accuracy is %.2f' %
        np.mean(scores)
    )
    logger.info('MVPA training and classification done')



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest


def test_R():
    from brainiak.factoranalysis.htfa import HTFA
    with pytest.raises(TypeError) as excinfo:
        HTFA()
    assert "missing 2 required positional arguments" in str(excinfo.value)


def test_X():
    from brainiak.factoranalysis.htfa import HTFA
    import numpy as np

    n_voxel = 100
    n_tr = 20
    K = 5
    max_global_iter = 3
    max_local_iter = 3
    max_voxel = n_voxel
    max_tr = n_tr

    R = []
    n_subj = 2
    for s in np.arange(n_subj):
        R.append(np.random.randint(2, high=102, size=(n_voxel, 3)))

    htfa = HTFA(
        K,
        n_subj=n_subj,
        max_global_iter=max_global_iter,
        max_local_iter=max_local_iter,
        max_voxel=max_voxel,
        max_tr=max_tr)

    X = np.random.rand(n_voxel, n_tr)
    # Check that does NOT run with wrong data type
    with pytest.raises(TypeError, match="Input data should be a list"):
        htfa.fit(X, R=R)

    X = []
    # Check that does NOT run with wrong array dimension
    with pytest.raises(ValueError,
                       match="Need at leat one subject to train the model"):
        htfa.fit(X, R=R)

    X = []
    X.append([1, 2, 3])
    # Check that does NOT run with wrong array dimension
    with pytest.raises(TypeError, match="data should be an array"):
        htfa.fit(X, R=R)

    X = []
    X.append(np.random.rand(n_voxel))
    # Check that does NOT run with wrong array dimension
    with pytest.raises(TypeError, match="subject data should be 2D array"):
        htfa.fit(X, R=R)

    X = []
    for s in np.arange(n_subj):
        X.append(np.random.rand(n_voxel, n_tr))
    R = np.random.randint(2, high=102, size=(n_voxel, 3))

    # Check that does NOT run with wrong data type
    with pytest.raises(TypeError, match="Coordinates should be a list"):
        htfa.fit(X, R=R)

    R = []
    R.append([1, 2, 3])
    # Check that does NOT run with wrong data type
    with pytest.raises(
            TypeError,
            match="Each scanner coordinate matrix should be an array"):
        htfa.fit(X, R=R)

    R = []
    R.append(np.random.rand(n_voxel))
    # Check that does NOT run with wrong array dimension
    with pytest.raises(
            TypeError,
            match="Each scanner coordinate matrix should be 2D array"):
        htfa.fit(X, R=R)

    R = []
    for s in np.arange(n_subj):
        R.append(np.random.rand(n_voxel - 1, 3))
    # Check that does NOT run with wrong array dimension
    with pytest.raises(
            TypeError,
            match=r"n_voxel should be the same in X\[idx\] and R\[idx\]"):
        htfa.fit(X, R=R)


@pytest.mark.mpiexec(n=2, timeout=100)
def test_can_run():
    import numpy as np
    from brainiak.factoranalysis.htfa import HTFA
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    n_voxel = 100
    n_tr = 20
    K = 5
    max_global_iter = 3
    max_local_iter = 3
    max_voxel = n_voxel
    max_tr = n_tr
    R = []
    n_subj = 2
    for s in np.arange(n_subj):
        R.append(np.random.randint(2, high=102, size=(n_voxel, 3)))
    my_R = []
    for idx in np.arange(n_subj):
        if idx % size == rank:
            my_R.append(R[idx])

    htfa = HTFA(
        K,
        n_subj=n_subj,
        max_global_iter=max_global_iter,
        max_local_iter=max_local_iter,
        max_voxel=max_voxel,
        max_tr=max_tr,
        verbose=True)
    assert htfa, "Invalid HTFA instance!"

    X = []
    for s in np.arange(n_subj):
        X.append(np.random.rand(n_voxel, n_tr))
    my_data = []
    for idx in np.arange(n_subj):
        if idx % size == rank:
            my_data.append(X[idx])

    if rank == 0:
        htfa.fit(my_data, R=my_R)
        assert True, "Root successfully running HTFA"
        assert htfa.global_prior_.shape[0] == htfa.prior_bcast_size,\
            "Invalid result of HTFA! (wrong # element in global_prior)"
        assert htfa.global_posterior_.shape[0] == htfa.prior_bcast_size,\
            "Invalid result of HTFA! (wrong # element in global_posterior)"

    else:
        htfa.fit(my_data, R=my_R)
        assert True, "worker successfully running HTFA"
        print(htfa.local_weights_.shape)
        assert htfa.local_weights_.shape[0] == n_tr * K,\
            "Invalid result of HTFA! (wrong # element in local_weights)"
        assert htfa.local_posterior_.shape[0] == htfa.prior_size,\
            "Invalid result of HTFA! (wrong # element in local_posterior)"


def test_dummy():
    """
    This is a dummy test to work around for the issue of pytest and
    pytest-mpiexec. See here the discussion of the same issue in
    pytest-forked:

    https://github.com/pytest-dev/pytest-forked/issues/67
    #issuecomment-1964718720

    """
    pass



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# Authors: Yida Wang
# (Intel Labs), 2016

cimport scipy.linalg.cython_blas as blas

def compute_self_corr_for_voxel_sel(py_trans_a, py_trans_b, py_m, py_n, py_k,
                                    py_alpha, py_a, py_lda, int py_start_voxel,
                                    py_b, py_ldb, py_beta, py_c, py_ldc,
                                    int py_start_epoch):
    """ use blas API sgemm wrapped by scipy to compute correlation

    This method is limited to process self-correlation.
    The blas APIs process matrices in column-major,
    but our matrices are in row-major,
    so we play the transpose trick here, i.e. A*B=(B^T*A^T)^T.
    The resulting matrix in shape [num_assigned_voxels, num_voxels]
    is stored in an alternate way to make sure that
    the correlation vectors of the same voxel stored continuously

    Parameters
    ----------
    py_trans_a: str
    do transpose or not for the first matrix A

    py_trans_b: str
    do transpose or not for the first matrix B

    py_m: int
    the row of the resulting matrix C
    in our case, is num_voxels

    py_n: int
    the column of the resulting matrix C
    in our case, is num_assigned_voxels

    py_k: int
    the collapsed dimension of the multiplying matrices
    i.e. the column of the first matrix after transpose if necessary
    the row of the second matrix after transpose if necessary

    py_alpha: float
    the weight applied to the first matrix A

    py_a: 2D array in shape [epoch_length, num_voxels]
    It is the activity data of an epoch, part 1 of the data to be
    correlated with. Note that py_a can point to the same location of py_b.

    py_lda: int
    the stride of the first matrix A

    py_start_voxel: int
    the starting voxel of assigned voxels
    used to locate the second matrix B

    py_b: 2D array in shape [epoch_length, num_voxels]
    It is the activity data of an epoch, part 2 of the data to be
    correlated with. Note that py_a can point to the same location of py_b.

    py_ldb: int
    the stride of the second matrix B

    py_beta: float
    the weight applied to the resulting matrix C

    py_c: 3D array in shape [num_selected_voxels, num_epochs, num_voxels]
    place to store the resulting correlation values

    py_ldc: int
    the stride of the resulting matrix
    in our case, num_voxels*num_epochs

    py_start_epoch: int
    the epoch over which the correlation is computed

    Returns
    -------
    py_c: 3D array in shape [num_selected_voxels, num_epochs, num_voxels]
    write the resulting correlation values in an alternate way
    for the processing epoch
    """
    cdef bytes by_trans_a=py_trans_a.encode()
    cdef bytes by_trans_b=py_trans_b.encode()
    cdef char* trans_a = by_trans_a
    cdef char* trans_b = by_trans_b
    cdef int M, N, K, lda, ldb, ldc
    M = py_m
    N = py_n
    K = py_k
    lda = py_lda
    ldb = py_ldb
    ldc = py_ldc
    cdef float alpha, beta
    alpha = py_alpha
    beta = py_beta
    cdef float[:, ::1] A
    A = py_a
    cdef float[:, ::1] B
    B = py_b
    cdef float[:, :, ::1] C
    C = py_c
    blas.sgemm(trans_a, trans_b, &M, &N, &K, &alpha, &A[0, 0], &lda,
               &B[0, py_start_voxel], &ldb, &beta, &C[0, py_start_epoch, 0], &ldc)

def compute_kernel_matrix(py_uplo, py_trans, py_n, py_k, py_alpha, py_a,
                          int py_start_voxel, py_lda,
                          py_beta, py_c, py_ldc):
    """ use blas API syrk wrapped by scipy to compute kernel matrix of SVM

    The blas APIs process matrices in column-major, but our matrices are
    in row-major, so we play the transpose trick here, i.e. A*B=(B^T*A^T)^T

    In SVM with linear kernel, the distance of two samples
    is essentially the dot product of them.
    Therefore, the kernel matrix can be obtained by matrix multiplication.
    Since the kernel matrix is symmetric, ssyrk is used,
    the other half of the matrix is assigned later.
    In our case, the dimension of samples is much larger than
    the number samples, so we proportionally shrink the values of
    the kernel matrix for getting more robust alpha values in SVM iteration.

    Parameters
    ----------
    py_uplo: str
    getting the upper or lower triangle of the matrix

    py_trans: str
    do transpose or not for the input matrix A

    py_n: int
    the row and column of the resulting matrix C
    in our case, is num_epochs

    py_k: int
    the collapsed dimension of the multiplying matrices
    i.e. the column of the first matrix after transpose if necessary
    the row of the second matrix after transpose if necessary
    in our case, is num_voxels

    py_alpha: float
    the weight applied to the input matrix A

    py_a: 3D array in shape [num_assigned_voxels, num_epochs, num_voxels]
    in our case the normalized correlation values of a voxel

    py_start_voxel: int
    the processed voxel
    used to locate the input matrix A

    py_lda: int
    the stride of the input matrix A

    py_beta: float
    the weight applied to the resulting matrix C

    py_c: 2D array in shape [num_epochs, num_epochs]
    place to store the resulting kernel matrix

    py_ldc: int
    the stride of the resulting matrix

    Returns
    -------
    py_c: 2D array in shape [num_epochs, num_epochs]
    write the resulting kernel_matrix
    for the processing voxel
    """
    cdef bytes by_uplo=py_uplo.encode()
    cdef bytes by_trans=py_trans.encode()
    cdef char* uplo = by_uplo
    cdef char* trans = by_trans
    cdef int N, K, lda, ldc
    N = py_n
    K = py_k
    lda = py_lda
    ldc = py_ldc
    cdef float alpha, beta
    alpha = py_alpha
    beta = py_beta
    cdef float[:, :, ::1] A
    A = py_a
    cdef float[:, ::1] C
    C = py_c
    blas.ssyrk(uplo, trans, &N, &K, &alpha, &A[py_start_voxel, 0, 0], &lda,
               &beta, &C[0, 0], &ldc)
    # complete the other half of the kernel matrix
    if py_uplo == 'L':
        for j in range(py_c.shape[0]):
            for k in range(j):
                py_c[j, k] = py_c[k, j]
    else:
        for j in range(py_c.shape[0]):
            for k in range(j):
                py_c[k, j] = py_c[j, k]

def compute_single_self_corr_syrk(py_uplo, py_trans, py_n, py_k,
                                  py_alpha, py_a, py_lda,
                                  py_beta, py_c, py_ldc,
                                  int py_start_sample):
    """ use blas API syrk wrapped by scipy to compute correlation matrix

    This is to compute the correlation between selected voxels for
    final training and classification. Since the resulting correlation
    matrix is symmetric, syrk is used. However, it looks like that in most
    cases, syrk performs much worse than gemm (the next function).
    Here we assume that the resulting matrix is stored in a compact way,
    i.e. py_ldc == py_n.

    Parameters
    ----------
    py_uplo: str
    getting the upper or lower triangle of the matrix

    py_trans: str
    do transpose or not for the input matrix A

    py_n: int
    the row and column of the resulting matrix C
    in our case, is num_selected_voxels

    py_k: int
    the collapsed dimension of the multiplying matrices
    i.e. the column of the first matrix after transpose if necessary
    the row of the second matrix after transpose if necessary
    in our case, is num_TRs

    py_alpha: float
    the weight applied to the input matrix A

    py_a: 2D array in shape [num_TRs, num_selected_voxels]
    in our case the normalized activity values

    py_lda: int
    the stride of the input matrix A

    py_beta: float
    the weight applied to the resulting matrix C

    py_c: 3D array
    in shape [num_samples, num_selected_voxels, num_selected_voxels]
    place to store the resulting kernel matrix

    py_ldc: int
    the stride of the resulting matrix

    py_start_sample: int
    the processed sample
    used to locate the resulting matrix C

    Returns
    -------
    py_c: 3D array
    in shape [num_samples, num_selected_voxels, num_selected_voxels]
    write the resulting correlation matrices
    for the processed sample
    """
    cdef bytes by_uplo=py_uplo.encode()
    cdef bytes by_trans=py_trans.encode()
    cdef char* uplo = by_uplo
    cdef char* trans = by_trans
    cdef int N, K, lda, ldc
    N = py_n
    K = py_k
    lda = py_lda
    ldc = py_ldc
    cdef float alpha, beta
    alpha = py_alpha
    beta = py_beta
    cdef float[:, ::1] A
    A = py_a
    cdef float[:, :, ::1] C
    C = py_c
    blas.ssyrk(uplo, trans, &N, &K, &alpha, &A[0, 0], &lda,
               &beta, &C[py_start_sample, 0, 0], &ldc)
    # complete the other half of the kernel matrix
    if py_uplo == 'L':
        for j in range(py_c.shape[1]):
            for k in range(j):
                py_c[py_start_sample, j, k] = py_c[py_start_sample, k, j]
    else:
        for j in range(py_c.shape[1]):
            for k in range(j):
                py_c[py_start_sample, k, j] = py_c[py_start_sample, j, k]

def compute_single_self_corr_gemm(py_trans_a, py_trans_b, py_m, py_n,
                                  py_k, py_alpha, py_a, py_lda,
                                  py_ldb, py_beta, py_c, py_ldc,
                                  int py_start_sample):
    """ use blas API gemm wrapped by scipy to compute correlation matrix

    This is to compute the correlation between selected voxels for
    final training and classification. Although the resulting correlation
    matrix is symmetric, in most cases, gemm performs better than syrk.
    Here we assume that the resulting matrix is stored in a compact way,
    i.e. py_ldc == py_n.

    Parameters
    ----------
    py_trans_a: str
    do transpose or not for the first matrix A

    py_trans_b: str
    do transpose or not for the first matrix B

    py_m: int
    the row of the resulting matrix C
    in our case, is num_selected_voxels

    py_n: int
    the column of the resulting matrix C
    in our case, is num_selected_voxels

    py_k: int
    the collapsed dimension of the multiplying matrices
    i.e. the column of the first matrix after transpose if necessary
    the row of the second matrix after transpose if necessary
    in our case, is num_TRs

    py_alpha: float
    the weight applied to the input matrix A

    py_a: 2D array in shape [num_TRs, num_selected_voxels]
    in our case the normalized activity values
    both multipliers are specified here as the same one

    py_lda: int
    the stride of the input matrix A

    py_ldb: int
    the stride of the input matrix B
    in our case, the same as py_lda

    py_beta: float
    the weight applied to the resulting matrix C

    py_c: 3D array
    in shape [num_samples, num_selected_voxels, num_selected_voxels]
    place to store the resulting kernel matrix

    py_ldc: int
    the stride of the resulting matrix

    py_start_sample: int
    the processed sample
    used to locate the resulting matrix C

    Returns
    -------
    py_c: 3D array
    in shape [num_samples, num_selected_voxels, num_selected_voxels]
    write the resulting correlation matrices
    for the processed sample
    """
    cdef bytes by_trans_a=py_trans_a.encode()
    cdef bytes by_trans_b=py_trans_b.encode()
    cdef char* trans_a = by_trans_a
    cdef char* trans_b = by_trans_b
    cdef int M, N, K, lda, ldb, ldc
    M = py_m
    N = py_n
    K = py_k
    lda = py_lda
    ldb = py_ldb
    ldc = py_ldc
    cdef float alpha, beta
    alpha = py_alpha
    beta = py_beta
    cdef float[:, ::1] A
    A = py_a
    cdef float[:, :, ::1] C
    C = py_c
    blas.sgemm(trans_a, trans_b, &M, &N, &K, &alpha, &A[0, 0], &lda,
               &A[0, 0], &ldb, &beta, &C[py_start_sample, 0, 0], &ldc)

def compute_corr_vectors(py_trans_a, py_trans_b, py_m, py_n,
                         py_k, py_alpha, py_a, py_lda,
                         py_b, py_ldb, py_beta, py_c, py_ldc,
                         int py_start_voxel,
                         int py_start_sample):
    """ use blas API gemm wrapped by scipy to construct a correlation vector

    The correlation vector is essentially correlation matrices computed
    from two activity matrices. It will be placed in the corresponding place
    of the resulting correlation data set.
    The blas APIs process matrices in column-major,
    but our matrices are in row-major, so we play the transpose trick here,
    i.e. A*B=(B^T*A^T)^T

    py_trans_a: str
    do transpose or not for the first matrix A

    py_trans_b: str
    do transpose or not for the first matrix B

    py_m: int
    the row of the resulting matrix C

    py_n: int
    the column of the resulting matrix C

    py_k: int
    the collapsed dimension of the multiplying matrices
    i.e. the column of the first matrix after transpose if necessary
    the row of the second matrix after transpose if necessary

    py_alpha: float
    the weight applied to the input matrix A

    py_a: 2D array

    py_lda: int
    the stride of the input matrix A

    py_b: 2D array

    py_ldb: int
    the stride of the input matrix B

    py_beta: float
    the weight applied to the resulting matrix C

    py_c: 2D array
    in shape [py_m, py_n] of column-major
    in fact it is
    in shape [py_n, py_m] of row-major

    py_ldc: int
    the stride of the resulting matrix

    py_start_voxel: int
    the starting voxel of assigned voxels
    used to locate the second matrix B

    py_start_sample: int
    the processed sample
    used to locate the resulting matrix C

    Returns
    -------
    py_c: 2D array
    in shape [py_m, py_n] of column-major
    write the resulting matrix to the place indicated by py_start_sample
    """
    cdef bytes by_trans_a=py_trans_a.encode()
    cdef bytes by_trans_b=py_trans_b.encode()
    cdef char* trans_a = by_trans_a
    cdef char* trans_b = by_trans_b
    cdef int M, N, K, lda, ldb, ldc
    M = py_m
    N = py_n
    K = py_k
    lda = py_lda
    ldb = py_ldb
    ldc = py_ldc
    cdef float alpha, beta
    alpha = py_alpha
    beta = py_beta
    cdef float[:, ::1] A
    A = py_a
    cdef float[:, ::1] B
    B = py_b
    cdef float[:, :, ::1] C
    C = py_c
    blas.sgemm(trans_a, trans_b, &M, &N, &K, &alpha, &A[0, 0], &lda,
               &B[0, py_start_voxel], &ldb, &beta, &C[py_start_sample, 0, 0], &ldc)

def compute_single_matrix_multiplication(py_trans_a, py_trans_b, py_m, py_n,
                                         py_k, py_alpha, py_a, py_lda,
                                         py_b, py_ldb, py_beta, py_c, py_ldc):
    """ use blas API gemm wrapped by scipy to do matrix multiplication

    This is to compute the matrix multiplication.
    The blas APIs process matrices in column-major,
    but our matrices are in row-major, so we play the transpose trick here,
    i.e. A*B=(B^T*A^T)^T

    Parameters
    ----------
    py_trans_a: str
    do transpose or not for the first matrix A

    py_trans_b: str
    do transpose or not for the first matrix B

    py_m: int
    the row of the resulting matrix C

    py_n: int
    the column of the resulting matrix C

    py_k: int
    the collapsed dimension of the multiplying matrices
    i.e. the column of the first matrix after transpose if necessary
    the row of the second matrix after transpose if necessary

    py_alpha: float
    the weight applied to the input matrix A

    py_a: 2D array

    py_lda: int
    the stride of the input matrix A

    py_b: 2D array

    py_ldb: int
    the stride of the input matrix B

    py_beta: float
    the weight applied to the resulting matrix C

    py_c: 2D array
    in shape [py_m, py_n] of column-major
    in fact it is
    in shape [py_n, py_m] of row-major

    py_ldc: int
    the stride of the resulting matrix

    Returns
    -------
    py_c: 2D array
    in shape [py_m, py_n] of column-major
    write the resulting matrix
    """
    cdef bytes by_trans_a=py_trans_a.encode()
    cdef bytes by_trans_b=py_trans_b.encode()
    cdef char* trans_a = by_trans_a
    cdef char* trans_b = by_trans_b
    cdef int M, N, K, lda, ldb, ldc
    M = py_m
    N = py_n
    K = py_k
    lda = py_lda
    ldb = py_ldb
    ldc = py_ldc
    cdef float alpha, beta
    alpha = py_alpha
    beta = py_beta
    cdef float[:, ::1] A
    A = py_a
    cdef float[:, ::1] B
    B = py_b
    cdef float[:, ::1] C
    C = py_c
    blas.sgemm(trans_a, trans_b, &M, &N, &K, &alpha, &A[0, 0], &lda,
               &B[0, 0], &ldb, &beta, &C[0, 0], &ldc)



#    Copyright 2016 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#             http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from multiprocessing import Pool
import numpy as np
from mpi4py import MPI
from scipy.spatial.distance import cityblock, euclidean

from ..utils.utils import usable_cpu_count

"""Distributed Searchlight
"""

__all__ = [
    "Ball",
    "Cube",
    "Diamond",
    "Searchlight",
    "Shape",
]


class Shape:
    """Shape

    Searchlight shape which is contained in a cube sized
    (2*rad+1,2*rad+1,2*rad+1)

    Attributes
    ----------

    mask_ : a 3D boolean numpy array of size (2*rad+1,2*rad+1,2*rad+1)
            which is set to True within the boundaries of the desired shape

    Parameters
    ----------

    rad: radius, in voxels, of the sphere inscribed in the
         searchlight cube, not counting the center voxel

    """

    def __init__(self, rad):
        self.rad = rad


class Cube(Shape):
    """Cube

    Searchlight shape which is a cube of size (2*rad+1,2*rad+1,2*rad+1)

    Parameters
    ----------

    rad: radius, in voxels, of the sphere inscribed in the
         searchlight cube, not counting the center voxel

    """
    def __init__(self, rad):
        super().__init__(rad)
        self.rad = rad
        self.mask_ = np.ones((2*rad+1, 2*rad+1, 2*rad+1), dtype=bool)


class Diamond(Shape):
    """Diamond

    Searchlight shape which is a diamond
    inscribed in a cube of size (2*rad+1,2*rad+1,2*rad+1).
    Any location in the cube which has a Manhattan distance of equal to or
    less than rad from the center point is set to True.

    Parameters
    ----------

    rad: radius, in voxels, of the sphere inscribed in the
         searchlight cube, not counting the center voxel

    """
    def __init__(self, rad):
        super().__init__(rad)
        self.mask_ = np.zeros((2*rad+1, 2*rad+1, 2*rad+1), dtype=bool)
        for r1 in range(2*self.rad+1):
            for r2 in range(2*self.rad+1):
                for r3 in range(2*self.rad+1):
                    if (cityblock((r1, r2, r3),
                                  (self.rad, self.rad, self.rad)) <= self.rad):
                        self.mask_[r1, r2, r3] = True


class Ball(Shape):
    """Ball

    Searchlight shape which is a ball
    inscribed in a cube of size (2*rad+1,2*rad+1,2*rad+1).
    Any location in the cube which has a Euclidean distance of equal to or
    less than rad from the center point is set to True.

    Parameters
    ----------

    rad: radius, in voxels, of the sphere inscribed in the
         searchlight cube, not counting the center voxel

    """
    def __init__(self, rad):
        super().__init__(rad)
        self.mask_ = np.zeros((2*rad+1, 2*rad+1, 2*rad+1), dtype=bool)
        for r1 in range(2*self.rad+1):
            for r2 in range(2*self.rad+1):
                for r3 in range(2*self.rad+1):
                    if (euclidean((r1, r2, r3),
                                  (self.rad, self.rad, self.rad)) <= self.rad):
                        self.mask_[r1, r2, r3] = True


class Searchlight:
    """Distributed Searchlight

    Run a user-defined function over each voxel in a multi-subject
    dataset.

    Optionally, users can define a block function which runs over
    larger portions of the volume called blocks.

    Parameters
    ----------

    sl_rad: radius, in voxels, of the sphere inscribed in the
               searchlight cube, not counting the center voxel

    max_blk_edge: max edge length, in voxels, of the 3D block

    shape: brainiak.searchlight.searchlight.Shape indicating the
    shape in voxels of the searchlight region

    min_active_voxels_proportion: float
        If a searchlight region does not have more than this minimum
        proportion of active voxels in the mask, it is not processed by the
        searchlight function. The mask used for the test is the
        intersection of the global (brain) mask and the `Shape` mask. The
        seed (central) voxel of the searchlight region is taken into
        consideration.

    pool_size: int
        Maximum number of processes running the function in parallel.
        If None, number of available hardware threads, considering cpusets
        restrictions. This value will be overridden by any pool_size passed
        to run or run_block_function by the user.

    """
    def __init__(self, sl_rad=1, max_blk_edge=10, shape=Cube,
                 min_active_voxels_proportion=0,
                 pool_size=None):
        assert sl_rad >= 0, 'sl_rad should not be negative'
        assert max_blk_edge > 0, 'max_blk_edge should be positive'
        self.sl_rad = sl_rad
        self.max_blk_edge = max_blk_edge
        self.min_active_voxels_proportion = min_active_voxels_proportion
        self.comm = MPI.COMM_WORLD
        self.shape = shape(sl_rad).mask_
        self.bcast_var = None
        self.pool_size = pool_size

    def _get_ownership(self, data):
        """Determine on which rank each subject currently resides

        Parameters
        ----------

        data: list of 4D arrays with subject data

        Returns
        -------

        list of ranks indicating the owner of each subject
        """
        rank = self.comm.rank

        B = [(rank, idx) for (idx, c) in enumerate(data) if c is not None]
        C = self.comm.allreduce(B)
        ownership = [None] * len(data)
        for c in C:
            ownership[c[1]] = c[0]
        return ownership

    def _get_blocks(self, mask):
        """Divide the volume into a set of blocks

        Ignore blocks that have no active voxels in the mask

        Parameters
        ----------

        mask: a boolean 3D array which is true at every active voxel

        Returns
        -------

        list of tuples containing block information:
           - a triple containing top left point of the block and
           - a triple containing the size in voxels of the block

        """
        blocks = []
        outerblk = self.max_blk_edge + 2*self.sl_rad
        for i in range(0, mask.shape[0], self.max_blk_edge):
            for j in range(0, mask.shape[1], self.max_blk_edge):
                for k in range(0, mask.shape[2], self.max_blk_edge):
                    block_shape = mask[i:i+outerblk,
                                       j:j+outerblk,
                                       k:k+outerblk
                                       ].shape
                    if np.any(
                        mask[i+self.sl_rad:i+block_shape[0]-self.sl_rad,
                             j+self.sl_rad:j+block_shape[1]-self.sl_rad,
                             k+self.sl_rad:k+block_shape[2]-self.sl_rad]):
                        blocks.append(((i, j, k), block_shape))
        return blocks

    def _get_block_data(self, mat, block):
        """Retrieve a block from a 3D or 4D volume

        Parameters
        ----------

        mat: a 3D or 4D volume

        block: a tuple containing block information:
          - a triple containing the lowest-coordinate voxel in the block
          - a triple containing the size in voxels of the block

        Returns
        -------

        In the case of a 3D array, a 3D subarray at the block location

        In the case of a 4D array, a 4D subarray at the block location,
        including the entire fourth dimension.
        """
        (pt, sz) = block
        if len(mat.shape) == 3:
            return mat[pt[0]:pt[0]+sz[0],
                       pt[1]:pt[1]+sz[1],
                       pt[2]:pt[2]+sz[2]].copy()
        elif len(mat.shape) == 4:
            return mat[pt[0]:pt[0]+sz[0],
                       pt[1]:pt[1]+sz[1],
                       pt[2]:pt[2]+sz[2],
                       :].copy()

    def _split_volume(self, mat, blocks):
        """Convert a volume into a list of block data

        Parameters
        ----------

        mat: A 3D or 4D array to be split

        blocks: a list of tuples containing block information:
          - a triple containing the top left point of the block and
          - a triple containing the size in voxels of the block


        Returns
        -------

        A list of the subarrays corresponding to each block

        """
        return [self._get_block_data(mat, block) for block in blocks]

    def _scatter_list(self, data, owner):
        """Distribute a list from one rank to other ranks in a cyclic manner

        Parameters
        ----------

        data: list of pickle-able data

        owner: rank that owns the data

        Returns
        -------

        A list containing the data in a cyclic layout across ranks

        """

        rank = self.comm.rank
        size = self.comm.size
        subject_submatrices = []
        nblocks = self.comm.bcast(len(data)
                                  if rank == owner else None, root=owner)

        # For each submatrix
        for idx in range(0, nblocks, size):
            padded = None
            extra = max(0, idx+size - nblocks)

            # Pad with "None" so scatter can go to all processes
            if data is not None:
                padded = data[idx:idx+size]
                if extra > 0:
                    padded = padded + [None]*extra

            # Scatter submatrices to all processes
            mytrans = self.comm.scatter(padded, root=owner)

            # Contribute submatrix to subject list
            if mytrans is not None:
                subject_submatrices += [mytrans]

        return subject_submatrices

    def distribute(self, subjects, mask):
        """Distribute data to MPI ranks

        Parameters
        ----------

        subjects : list of 4D arrays containing data for one or more subjects.
              Each entry of the list must be present on at most one rank,
              and the other ranks contain a "None" at this list location.

              For example, for 3 ranks you may lay out the data in the
              following manner:

              Rank 0: [Subj0, None, None]
              Rank 1: [None, Subj1, None]
              Rank 2: [None, None, Subj2]

              Or alternatively, you may lay out the data in this manner:

              Rank 0: [Subj0, Subj1, Subj2]
              Rank 1: [None, None, None]
              Rank 2: [None, None, None]

        mask: 3D array with "True" entries at active vertices

        """
        if mask.ndim != 3:
            raise ValueError('mask should be a 3D array')

        for (idx, subj) in enumerate(subjects):
            if subj is not None:
                if subj.ndim != 4:
                    raise ValueError('subjects[{}] must be 4D'.format(idx))

        self.mask = mask
        rank = self.comm.rank

        # Get/set ownership
        ownership = self._get_ownership(subjects)
        all_blocks = self._get_blocks(mask) if rank == 0 else None
        all_blocks = self.comm.bcast(all_blocks)

        # Divide data and mask
        splitsubj = [self._split_volume(s, all_blocks)
                     if s is not None else None
                     for s in subjects]
        submasks = self._split_volume(mask, all_blocks)

        # Scatter points, data, and mask
        self.blocks = self._scatter_list(all_blocks, 0)
        self.submasks = self._scatter_list(submasks,  0)
        self.subproblems = [self._scatter_list(s, ownership[s_idx])
                            for (s_idx, s) in enumerate(splitsubj)]

    def broadcast(self, bcast_var):
        """Distribute data to processes

        Parameters
        ----------

        bcast_var:    shared data which is broadcast to all processes

        """

        self.bcast_var = self.comm.bcast(bcast_var)

    def run_block_function(self, block_fn, extra_block_fn_params=None,
                           pool_size=None):
        """Perform a function for each block in a volume.

        Parameters
        ----------

        block_fn: function to apply to each block:

                Parameters

                data: list of 4D arrays containing subset of subject data,
                      which is padded with sl_rad voxels.

                mask: 3D array containing subset of mask data

                sl_rad: radius, in voxels, of the sphere inscribed in the
                cube

                bcast_var: shared data which is broadcast to all processes

                extra_params: extra parameters


                Returns

                3D array which is the same size as the mask
                input with padding removed

        extra_block_fn_params: tuple
            Extra parameters to pass to the block function

        pool_size: int
            Maximum number of processes running the block function in parallel.
            If None, number of available hardware threads, considering cpusets
            restrictions.
        """
        rank = self.comm.rank

        results = []
        usable_cpus = usable_cpu_count()

        # The user specified a pool_size on the searchlight object, so use that
        if pool_size is None and self.pool_size is not None:
            pool_size = self.pool_size

        if pool_size is None:
            processes = usable_cpus
        else:
            processes = min(pool_size, usable_cpus)

        if processes > 1:
            with Pool(processes) as pool:
                for idx, block in enumerate(self.blocks):
                    result = pool.apply_async(
                        block_fn,
                        ([subproblem[idx] for subproblem in self.subproblems],
                         self.submasks[idx],
                         self.sl_rad,
                         self.bcast_var,
                         extra_block_fn_params))
                    results.append((block[0], result))
                local_outputs = [(result[0], result[1].get())
                                 for result in results]
        else:
            # If we only are using one CPU core, no need to create a Pool,
            # cause an underlying fork(), and send the data to that process.
            # Just do it here in serial. This  will save copying the memory
            # and will stop a fork() which can cause problems in some MPI
            # implementations.
            for idx, block in enumerate(self.blocks):
                subprob_list = [subproblem[idx]
                                for subproblem in self.subproblems]
                result = block_fn(
                            subprob_list,
                            self.submasks[idx],
                            self.sl_rad,
                            self.bcast_var,
                            extra_block_fn_params)
                results.append((block[0], result))
            local_outputs = [(result[0], result[1]) for result in results]

        # Collect results
        global_outputs = self.comm.gather(local_outputs)

        # Coalesce results
        outmat = np.empty(self.mask.shape, dtype=object)
        if rank == 0:
            for go_rank in global_outputs:
                for (pt, mat) in go_rank:
                    coords = np.s_[
                        pt[0]+self.sl_rad:pt[0]+self.sl_rad+mat.shape[0],
                        pt[1]+self.sl_rad:pt[1]+self.sl_rad+mat.shape[1],
                        pt[2]+self.sl_rad:pt[2]+self.sl_rad+mat.shape[2]
                    ]
                    outmat[coords] = mat
        return outmat

    def run_searchlight(self, voxel_fn, pool_size=None):
        """Perform a function at each voxel which is set to True in the
        user-provided mask. The mask passed to the searchlight function will be
        further masked by the user-provided searchlight shape.

        Parameters
        ----------

        voxel_fn: function to apply at each voxel

            Must be `serializeable using pickle
            <https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled>`_.

                Parameters

                subj: list of 4D arrays containing subset of subject data

                mask: 3D array containing subset of mask data

                sl_rad: radius, in voxels, of the sphere inscribed in the
                cube

                bcast_var: shared data which is broadcast to all processes


                Returns

                Value of any pickle-able type

        Returns
        -------

        A volume which is the same size as the mask, however a number of voxels
        equal to the searchlight radius has been removed from each border of
        the volume. This volume contains the values returned from the
        searchlight function at each voxel which was set to True in the mask,
        and None elsewhere.

        """
        # The user specified a pool_size on the searchlight object, so use that
        if pool_size is None and self.pool_size is not None:
            pool_size = self.pool_size

        extra_block_fn_params = (voxel_fn, self.shape,
                                 self.min_active_voxels_proportion)
        block_fn_result = self.run_block_function(_singlenode_searchlight,
                                                  extra_block_fn_params,
                                                  pool_size)
        return block_fn_result


def _singlenode_searchlight(data, msk, mysl_rad, bcast_var, extra_params):
    """Run searchlight function on block data in parallel.

    `extra_params` contains:

    - Searchlight function.
    - `Shape` mask.
    - Minimum active voxels proportion required to run the searchlight
      function.
    """

    voxel_fn = extra_params[0]
    shape_mask = extra_params[1]
    min_active_voxels_proportion = extra_params[2]
    outmat = np.empty(msk.shape, dtype=object)
    if mysl_rad > 0:
        outmat = outmat[mysl_rad:-mysl_rad,
                        mysl_rad:-mysl_rad,
                        mysl_rad:-mysl_rad]
    for i in range(0, outmat.shape[0]):
        for j in range(0, outmat.shape[1]):
            for k in range(0, outmat.shape[2]):
                if msk[i+mysl_rad, j+mysl_rad, k+mysl_rad]:
                    searchlight_slice = np.s_[
                        i:i+2*mysl_rad+1,
                        j:j+2*mysl_rad+1,
                        k:k+2*mysl_rad+1]
                    voxel_fn_mask = msk[searchlight_slice] * shape_mask
                    if (min_active_voxels_proportion == 0
                        or np.count_nonzero(voxel_fn_mask) / voxel_fn_mask.size
                            > min_active_voxels_proportion):
                        outmat[i, j, k] = voxel_fn(
                            [subject[searchlight_slice] for subject in data],
                            msk[searchlight_slice] * shape_mask,
                            mysl_rad,
                            bcast_var)
    return outmat



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
import sys
import math
import requests 
import scipy.io
import numpy as np
from mpi4py import MPI
from subprocess import call
from scipy.stats import stats
from sklearn import model_selection 
from sklearn.metrics import mean_squared_error

def recon_err(data, F, W):
    """Calcuate reconstruction error

    Parameters
    ----------

    data : 2D array 
        True data to recover.

    F : 2D array 
        HTFA factor matrix.

    W : 2D array 
        HTFA weight matrix.
   

    Returns
    -------
    float 
        Returns root mean squared reconstruction error.

    """

    recon = F.dot(W).ravel()
    err = mean_squared_error(
        data.ravel(),
        recon,
        multioutput='uniform_average')
    return math.sqrt(err)

def get_train_err(htfa, data, F):
    """Calcuate training error

    Parameters
    ----------

    htfa : HTFA
        An instance of HTFA, factor anaysis class in BrainIAK.
   
    data : 2D array 
        Input data to HTFA.

    F : 2D array 
        HTFA factor matrix.
        
    Returns
    -------
    float 
        Returns root mean squared error on training.

    """

    W = htfa.get_weights(data, F)
    return recon_err(data, F, W)

def get_test_err(htfa, test_weight_data, test_recon_data,
        test_weight_R, test_recon_R, centers, widths):

    """Calcuate test error
    
    Parameters
    ----------

    htfa : HTFA
        An instance of HTFA, factor anaysis class in BrainIAK.
   
    test_weigth_data : 2D array 
        Data used for testing weights.
    
    test_recon_data : 2D array 
        Data used for testing reconstruction error.

    test_weigth_R : 2D array 
        Coordinate matrix used for testing weights.
    
    test_recon_R : 2D array 
        Coordinate matrix used for testing reconstruction error.

    centers : 2D array 
        Center matrix of HTFA factors.
    
    widths : 1D array 
        Width matrix of HTFA factors.
        
    Returns
    -------
    float 
        Returns root mean squared error on test.

    """

    # calculate F on test_weight_R, based on trained centers/widths
    unique_R, inds = htfa.get_unique_R(test_weight_R)
    F = htfa.get_factors(unique_R,
                         inds,
                         centers,
                         widths)
    # calculate weights on test_weight_data
    W = htfa.get_weights(test_weight_data, F)

    # calculate F on final test_recon_data
    unique_R, inds = htfa.get_unique_R(test_recon_R)
    F = htfa.get_factors(unique_R,
                         inds,
                         centers,
                         widths)
    return recon_err(test_recon_data, F, W)

n_subj = 2
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()
group_id = int(rank/n_subj)
n_group = math.ceil(size/n_subj)
htfa_comm = comm.Split(group_id, rank)
htfa_rank = htfa_comm.Get_rank()
htfa_size = htfa_comm.Get_size()

if rank == 0:
    import logging
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)

data_dir = os.path.join(os.getcwd(), 'data')
if rank == 0 and not os.path.exists(data_dir):
    os.makedirs(data_dir)

url = []
url.append(' https://www.dropbox.com/s/r5s9tg4ekxzbrco/s0.mat?dl=0')
url.append(' https://www.dropbox.com/s/39tr01m76vxwaqa/s1.mat?dl=0')

for idx in range(n_subj):
    if idx % size == rank:
        file_name = os.path.join(data_dir, 's' + str(idx) + '.mat')
        #check if file has already been downloaded
        if not os.path.exists(file_name):
            #check if URL exists
            ret = requests.head(url[idx])
            if ret.status_code == 200:
                #download data
                cmd = 'curl --location -o ' + file_name + url[idx]
                try:
                    retcode = call(cmd, shell=True)
                    if retcode < 0:
                        print("File download was terminated by signal", -retcode, file=sys.stderr)
                    else:
                        print("File download returned", retcode, file=sys.stderr)
                except OSError as e:
                    print("File download failed:", e, file=sys.stderr)
            else:
                 print("File s%d.mat does not exist!\n"%idx)

comm.Barrier()

#get fMRI data and scanner RAS coordinates
data = []
R = []
mapping = {}
n_local_subj = 0
for idx in range(n_subj):
    if idx % htfa_size == htfa_rank:
        file_name = os.path.join(data_dir, 's' + str(idx) + '.mat')
        all_data = scipy.io.loadmat(file_name)
        bold = all_data['data']
        # z-score the data
        bold = stats.zscore(bold, axis=1, ddof=1)
        data.append(bold)
        R.append(all_data['R'])
        mapping[str(n_local_subj)] = idx
        n_local_subj += 1


min_K = 3
max_K = 6
n_K = 2
Ks  = np.linspace(min_K, max_K, n_K, endpoint=True).astype(int)
n_splits = 3
# recon_err in shape n_splits*n_K
test_recon_errs = np.zeros((n_subj, n_splits, n_K))
tmp_test_recon_errs = np.zeros((n_subj, n_splits, n_K))
train_recon_errs = np.zeros((n_subj, n_splits, n_K))
tmp_train_recon_errs = np.zeros((n_subj, n_splits, n_K))

local_size = math.ceil(n_subj/size)
if n_local_subj > 0:
    from brainiak.factoranalysis.htfa import HTFA
    n_voxel, n_tr = data[0].shape
    n_dim = R[0].shape[1]
    test_size = 0.3
    rnd_seed_voxel = 30000
    rnd_seed_tr = 3000
    tr_solver = 'exact'
    nlss_method = 'dogbox'
    nlss_loss = 'linear'
    upper_ratio = 1.8
    lower_ratio = 0.1
    voxel_ratio = 0.25
    tr_ratio = 0.1
    max_voxel = 2000
    max_tr = 200
    max_sample_voxel = min(max_voxel,
                          int(voxel_ratio * n_voxel))
    max_sample_tr = min(max_tr, int(tr_ratio * n_tr))

    #split voxel and TR for two-level cross validation
    ss_voxel = model_selection.ShuffleSplit(
                                     n_splits=n_splits,
                                     test_size=test_size,
                                     random_state=rnd_seed_voxel)
    voxel_indices = np.arange(n_voxel)
    ss_voxel.get_n_splits(voxel_indices)

    ss_tr = model_selection.ShuffleSplit(
                                     n_splits=n_splits,
                                     test_size=test_size,
                                     random_state=rnd_seed_tr)
    tr_indices = np.arange(n_tr)
    ss_tr.get_n_splits(tr_indices)

    train_voxels = []
    test_voxels = []
    train_trs = []
    test_trs = []
    for train_index, test_index in ss_voxel.split(voxel_indices):
        train_voxels.append(train_index)
        test_voxels.append(test_index)

    for train_index, test_index in ss_tr.split(tr_indices):
        train_trs.append(train_index)
        test_trs.append(test_index)


    for p in range(n_splits):
        for idx in range(n_K):
            index = p*n_K + idx
            if index % n_group == group_id:
                #split data and R
                train_voxel_indices = train_voxels[p]
                test_voxel_indices = test_voxels[p]
                train_tr_indices = train_trs[p]
                test_tr_indices = test_trs[p]

                train_data = []
                total_test_data = []
                test_weight_data = []
                test_recon_data = []
                test_weight_R = []
                test_recon_R = []
                for s in range(n_local_subj):
                    train_data.append(data[s][:, train_tr_indices])
                    total_test_data.append(data[s][:, test_tr_indices])
                    test_weight_data.append(
                            total_test_data[s][train_voxel_indices, :])
                    test_recon_data.append(
                            total_test_data[s][test_voxel_indices, :])
                    test_weight_R.append(R[s][train_voxel_indices])
                    test_recon_R.append(R[s][test_voxel_indices])

                htfa = HTFA(K=Ks[idx],
                        max_global_iter=5,
                        max_local_iter=2,
                        n_subj=n_subj,
                        nlss_method=nlss_method,
                        nlss_loss=nlss_loss,
                        tr_solver=tr_solver,
                        upper_ratio=upper_ratio,
                        lower_ratio=lower_ratio,
                        max_tr=max_sample_tr,
                        max_voxel=max_sample_voxel,
                        comm=htfa_comm,
                        verbose=True)
                htfa.fit(train_data, R)

                for s in range(n_local_subj):
                    #get posterior for each subject
                    subj_idx = mapping[str(s)]
                    start_idx = s * htfa.prior_size
                    end_idx = (s + 1) * htfa.prior_size
                    local_posteiror = htfa.local_posterior_[start_idx:end_idx]
                    local_centers = htfa.get_centers(local_posteiror)
                    local_widths = htfa.get_widths(local_posteiror)

                    htfa.n_dim = n_dim
                    htfa.cov_vec_size = np.sum(np.arange(htfa.n_dim) + 1)
                    htfa.map_offset = htfa.get_map_offset()
                    #training happens on all voxels, but part of TRs
                    unique_R_all, inds_all = htfa.get_unique_R(R[s])
                    train_F = htfa.get_factors(unique_R_all,
                                             inds_all,
                                             local_centers,
                                             local_widths)

                    #calculate train_recon_err
                    tmp_train_recon_errs[subj_idx, p,idx] = get_train_err(htfa,
                                                         train_data[s],
                                                         train_F)

                    #calculate weights on test_weight_data, test_recon_err on test_recon_data
                    tmp_test_recon_errs[subj_idx, p,idx] = get_test_err(htfa,
                                                        test_weight_data[s],
                                                        test_recon_data[s],
                                                        test_weight_R[s],
                                                        test_recon_R[s],
                                                        local_centers,
                                                        local_widths)

comm.Reduce(tmp_test_recon_errs, test_recon_errs, op=MPI.SUM)
comm.Reduce(tmp_train_recon_errs, train_recon_errs, op=MPI.SUM)

if rank == 0:
    errs = train_recon_errs.reshape(n_subj * n_splits, n_K)
    mean_errs = np.average(errs, axis=0)
    print("train error on each K is\n")
    print(mean_errs)
    errs = test_recon_errs.reshape(n_subj * n_splits, n_K)
    mean_errs = np.average(errs, axis=0)
    print("test error on each K is\n")
    print(mean_errs)
    best_idx = np.argmin(mean_errs)
    print("best K for test recon is %d " % (Ks[best_idx]))



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Example for using hyperparameter optimization (hpo) package.

In this example, we will try to optimize a function of
2 variables (branin) using both hpo and grid search. 

"""

import brainiak.hyperparamopt.hpo as hpo
import numpy as np
import scipy.stats as st
import matplotlib.pyplot as plt

# Branin is the function we want to minimize.
# It is a function of 2 variables.
# In the range x1 in [-5, 10] and x2 in [0, 15],
# this function has 2 local minima and 1 global minima.
# Global minima of -16.6 at (-3.7, 13.7). 
# This is the modified version (Branin-Hoo) of the standard branin function.
# If you want the standard version (which has 3 global minima),
# you can omit the "+5*x1" term at the end
# For more details, see http://www.sfu.ca/~ssurjano/branin.html
def branin(x1, x2):
    a = 1.0
    b = 5.1/(4*np.pi*np.pi)
    c = 5.0/np.pi
    r = 6.0
    s = 10.0
    t = 1.0/(8*np.pi)
    return a*((x2 - b*x1*x1 + c*x1 - r)**2) + s*(1-t)*np.cos(x1) + s + 5*x1

# This is a wrapper around branin that takes in a dictionary
def branin_wrapper(args):
    x1 = args['x1']
    x2 = args['x2']
    return branin(x1,x2)

# Define ranges for the two variables
x1lo = -5
x1hi = 10
x2lo = 0
x2hi = 15

##############################
# Optimization through hpo
##############################

# Define a space for hpo to use
# The space needs to define
# 1. Name of the variables
# 2. Default samplers for the variables (use scipy.stats objects)
# 3. lo and hi ranges for the variables (will use -inf, inf if not specified)
space = {'x1':{'dist': st.uniform(x1lo, x1hi-x1lo), 'lo':x1lo, 'hi':x1hi},
         'x2':{'dist': st.uniform(x2lo, x2hi-x2lo), 'lo':x2lo, 'hi':x2hi}}

# The trials object is just a list that stores the samples generated and the
# corresponding function values at those sample points.
trials = []

# Maximum number of samples that will be generated.
# This is the maximum number of function evaluations that will be performed.
n_hpo_samples = 100

# Call the fmin function that does the optimization.
# The function to be optimized should take in a dictionary. You will probably
# need to wrap your function to do this (see branin() and branin_wrapper()).
# You can pass in a non-empty trials object as well e.g. from a previous
# fmin run. We just append to the trials object and will use existing data
# in our optimization.
print("Starting optimization through hpo")
best = hpo.fmin(loss_fn=branin_wrapper, space=space,
                max_evals=n_hpo_samples, trials=trials)

# Print out the best value obtained through HPO
print("Best obtained through HPO (", n_hpo_samples, " samples) = ",
       best['x1'], best['x2'], "; min value = ", best['loss'])

#####################################
# Optimization through grid search
#####################################

# Divide the space into a uniform grid (meshgrid)
n = 200
x1 = np.linspace(x1lo, x1hi, n)
x2 = np.linspace(x2lo, x2hi, n)
x1_grid, x2_grid = np.meshgrid(x1, x2)

# Calculate the function values along the grid
print("Starting optimization through grid search")
z = branin(x1_grid, x2_grid)

# Print out the best value obtained through grid search
print("Best obtained through grid search (", n*n, " samples) = ",
       x1_grid.flatten()[z.argmin()], x2_grid.flatten()[z.argmin()],
       "; min value = ", z.min())

########
# Plots
########

# Convert trials object data into numpy arrays
x1 = np.array([tr['x1'] for tr in trials])
x2 = np.array([tr['x2'] for tr in trials])
y = np.array([tr['loss'] for tr in trials])

# Plot the function contour using the grid search data
h = (z.max()-z.min())/25
plt.contour(x1_grid, x2_grid, z, levels=np.linspace(z.min()-h, z.max(), 26))

# Mark the points that were sampled through HPO
plt.scatter(x1, x2, s=10, color='r', label='HPO Samples')

# Mark the best points obtained through both methods
plt.scatter(best['x1'], best['x2'], s=30, color='b', label='Best HPO')
plt.scatter(x1_grid.flatten()[z.argmin()], x2_grid.flatten()[z.argmin()],
            s=30, color='g', label='Best grid search')

# Labels
plt.xlabel('x1')
plt.ylabel('x2')
plt.title('Hyperparameter optimization using HPO (Branin function)')
plt.legend()
plt.show()




#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest


def test_tfa():
    from brainiak.factoranalysis.tfa import TFA
    import numpy as np

    n_voxel = 100
    n_tr = 20
    K = 5
    max_iter = 5
    max_num_voxel = n_voxel
    max_num_tr = n_tr
    tfa = TFA(
        K=K,
        max_iter=max_iter,
        verbose=True,
        max_num_voxel=max_num_voxel,
        max_num_tr=max_num_tr)
    assert tfa, "Invalid TFA instance!"

    R = np.random.randint(2, high=102, size=(n_voxel, 3))
    X = [1, 2, 3]
    # Check that does NOT run with wrong data type
    with pytest.raises(TypeError) as excinfo:
        tfa.fit(X, R=R)
    assert "Input data should be an array" in str(excinfo.value)

    X = np.random.rand(n_voxel)
    # Check that does NOT run with wrong array dimension
    with pytest.raises(TypeError) as excinfo:
        tfa.fit(X, R=R)
    assert "Input data should be 2D array" in str(excinfo.value)

    X = np.random.rand(n_voxel, n_tr)
    R = [1, 2, 3]
    # Check that does NOT run with wrong data type
    with pytest.raises(TypeError) as excinfo:
        tfa.fit(X, R=R)
    assert "coordinate matrix should be an array" in str(excinfo.value)

    R = np.random.rand(n_voxel)
    # Check that does NOT run with wrong array dimension
    with pytest.raises(TypeError) as excinfo:
        tfa.fit(X, R=R)
    assert "coordinate matrix should be 2D array" in str(excinfo.value)

    R = np.random.randint(2, high=102, size=(n_voxel - 1, 3))
    # Check that does NOT run if n_voxel in X and R does not match
    with pytest.raises(TypeError) as excinfo:
        tfa.fit(X, R=R)
    assert "The number of voxels should be the same in X and R" in str(
        excinfo.value)

    R = np.random.randint(2, high=102, size=(n_voxel, 3))
    tfa.fit(X, R=R)
    assert True, "Success running TFA with one subject!"
    posterior_size = K * (tfa.n_dim + 1)
    assert tfa.local_posterior_.shape[
        0] == posterior_size,\
        "Invalid result of TFA! (wrong # element in local_posterior)"

    weight_method = 'ols'
    tfa = TFA(
        weight_method=weight_method,
        K=K,
        max_iter=max_iter,
        verbose=True,
        max_num_voxel=max_num_voxel,
        max_num_tr=max_num_tr)
    assert tfa, "Invalid TFA instance!"

    X = np.random.rand(n_voxel, n_tr)
    tfa.fit(X, R=R)
    assert True, "Success running TFA with one subject!"

    template_prior, _, _ = tfa.get_template(R)
    tfa.set_K(K)
    tfa.set_seed(200)
    tfa.fit(X, R=R, template_prior=template_prior)
    assert True, "Success running TFA with one subject and template prior!"
    assert tfa.local_posterior_.shape[
        0] == posterior_size,\
        "Invalid result of TFA! (wrong # element in local_posterior)"

    weight_method = 'odd'
    tfa = TFA(
        weight_method=weight_method,
        K=K,
        max_iter=max_iter,
        verbose=True,
        max_num_voxel=max_num_voxel,
        max_num_tr=max_num_tr)
    with pytest.raises(ValueError) as excinfo:
        tfa.fit(X, R=R)
    assert "'rr' and 'ols' are accepted as weight_method!" in str(
        excinfo.value)



#  Copyright 2016 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from pathlib import Path

import numpy as np

from brainiak.fcma.preprocessing import (prepare_fcma_data, prepare_mvpa_data,
                                         prepare_searchlight_mvpa_data)
from brainiak import io

data_dir = Path(__file__).parents[1] / 'io' / 'data'
expected_dir = Path(__file__).parent / 'data'
suffix = 'bet.nii.gz'
mask_file = data_dir / 'mask.nii.gz'
epoch_file = data_dir / 'epoch_labels.npy'
expected_labels = np.array([0, 1, 0, 1])


def test_prepare_fcma_data():
    images = io.load_images_from_dir(data_dir, suffix=suffix)
    mask = io.load_boolean_mask(mask_file)
    conditions = io.load_labels(epoch_file)
    raw_data, _, labels = prepare_fcma_data(images, conditions, mask)
    expected_raw_data = np.load(expected_dir / 'expected_raw_data.npy')
    assert len(raw_data) == len(expected_raw_data), \
        'numbers of epochs do not match in test_prepare_fcma_data'
    for idx in range(len(raw_data)):
        assert np.allclose(raw_data[idx], expected_raw_data[idx]), \
            'raw data do not match in test_prepare_fcma_data'
    assert np.array_equal(labels, expected_labels), \
        'the labels do not match in test_prepare_fcma_data'
    from brainiak.fcma.preprocessing import RandomType
    images = io.load_images_from_dir(data_dir, suffix=suffix)
    random_raw_data, _, _ = prepare_fcma_data(images, conditions, mask,
                                              random=RandomType.REPRODUCIBLE)
    assert len(random_raw_data) == len(expected_raw_data), \
        'numbers of epochs do not match in test_prepare_fcma_data'
    images = io.load_images_from_dir(data_dir, suffix=suffix)
    random_raw_data, _, _ = prepare_fcma_data(images, conditions, mask,
                                              random=RandomType.UNREPRODUCIBLE)
    assert len(random_raw_data) == len(expected_raw_data), \
        'numbers of epochs do not match in test_prepare_fcma_data'


def test_prepare_mvpa_data():
    images = io.load_images_from_dir(data_dir, suffix=suffix)
    mask = io.load_boolean_mask(mask_file)
    conditions = io.load_labels(epoch_file)
    processed_data, labels = prepare_mvpa_data(images, conditions, mask)
    expected_processed_data = np.load(expected_dir
                                      / 'expected_processed_data.npy')
    assert len(processed_data) == len(expected_processed_data), \
        'numbers of epochs do not match in test_prepare_mvpa_data'
    for idx in range(len(processed_data)):
        assert np.allclose(processed_data[idx],
                           expected_processed_data[idx]), (
            'raw data do not match in test_prepare_mvpa_data')
    assert np.array_equal(labels, expected_labels), \
        'the labels do not match in test_prepare_mvpa_data'


def test_prepare_searchlight_mvpa_data():
    images = io.load_images_from_dir(data_dir, suffix=suffix)
    conditions = io.load_labels(epoch_file)
    processed_data, labels = prepare_searchlight_mvpa_data(images,
                                                           conditions)
    expected_searchlight_processed_data = np.load(
        expected_dir / 'expected_searchlight_processed_data.npy')
    for idx in range(len(processed_data)):
        assert np.allclose(processed_data[idx],
                           expected_searchlight_processed_data[idx]), (
            'raw data do not match in test_prepare_searchlight_mvpa_data')
    assert np.array_equal(labels, expected_labels), \
        'the labels do not match in test_prepare_searchlight_mvpa_data'
    from brainiak.fcma.preprocessing import RandomType
    images = io.load_images_from_dir(data_dir, suffix=suffix)
    random_processed_data, _ = prepare_searchlight_mvpa_data(
        images,
        conditions,
        random=RandomType.REPRODUCIBLE)
    assert (len(random_processed_data)
            == len(expected_searchlight_processed_data)), (
        'numbers of epochs do not match in test_prepare_searchlight_mvpa_data')
    images = io.load_images_from_dir(data_dir, suffix=suffix)
    random_processed_data, _ = prepare_searchlight_mvpa_data(
        images,
        conditions,
        random=RandomType.UNREPRODUCIBLE)
    assert (len(random_processed_data)
            == len(expected_searchlight_processed_data)), (
        'numbers of epochs do not match in test_prepare_searchlight_mvpa_data')


if __name__ == '__main__':
    test_prepare_fcma_data()
    test_prepare_mvpa_data()
    test_prepare_searchlight_mvpa_data()
